"""

.. autoclass:: DataIO
   :members:

"""

import os, shutil
import json
from collections import OrderedDict
import numpy as np
import pandas as pd
from urllib.request import urlretrieve
import pickle
import distutils.version


import sklearn.metrics

from .version import version as tridesclous_version
from .datasource import data_source_classes
from .iotools import ArrayCollection
from .tools import download_probe, create_prb_file_from_dict, fix_prb_file_py2
from .waveformtools import extract_chunks

from .export import export_list, export_dict

_signal_types = ['initial', 'processed']





class DataIO:
    """
    Class to acces the dataset (raw data, processed, catalogue, 
    spikes) in read/write mode.
    
    All operations on the dataset are done througth that class.
    
    The dataio :
      * work in a path. Almost everything is persistent.
      * needed by CatalogueConstructor and Peeler
      * have a datasource member that access raw data
      * store/load processed signals
      * store/load spikes
      * store/load the catalogue
      * deal with sevral channel groups given a PRB file
      * deal with several segment of recording (aka several files for raw data)
      * export the results (labeled spikes) to differents format
      
    
    The underlying data storage is a simple tree on the file system.
    Everything is organised as simple as possible in sub folder 
    (by channel group then segment index).
    In each folder:
      * arrays.json describe the list of numpy array (name, dtype, shape)
      * XXX.raw are the raw numpy arrays and load with a simple memmap.
      * some array are struct arrays (aka array of struct)
      
    The datasource system is based on neo.rawio so all format in neo.rawio are
    available in tridesclous. neo.rawio is able to read chunk of signals indexed 
    on time axis and channel axis.
    
    The raw dataset do not need to be inside the working directory but can be somewhere outside.
    The info.json describe the link to the *datasource* (raw data)
    
    Many raw dataset are saved by the device with an underlying int16.
    DataIO save the processed signals as float32 by default. So if
    you have a 10Go raw dataset tridesclous will need at least 20 Go more for storage
    of the processed signals.
    
    
    **Usage**::
    
        # initialize a directory
        dataio = DataIO(dirname='/path/to/a/working/dir')
        
        # set a data source
        filenames = ['file1.raw', 'file2.raw']
        dataio.dataio.set_data_source(type='RawData', filenames=filenames, 
                                    sample_rate=10000, total_channel=16, dtype='int16')
        
        # set a PRB file
        dataio.set_probe_file('/path/to/a/file.prb')
        # or dowload it
        dataio.download_probe('kampff_128', origin='spyking-circus')
        
        # check lenght and channel groups
        print(dataio)

    
    """
    
    @staticmethod
    def check_initialized(dirname):
        if not os.path.exists(dirname):
            return False

        info_filename = os.path.join(dirname, 'info.json')
        if not os.path.exists(info_filename):
            return False
        
        return True
    
    def __init__(self, dirname='test'):
        self.dirname = dirname
        if not os.path.exists(dirname):
            os.mkdir(dirname)
        
        self.info_filename = os.path.join(self.dirname, 'info.json')
        if not os.path.exists(self.info_filename):
            #first init
            self.info = {}
            self.info['tridesclous_version'] = tridesclous_version
            self.flush_info()
            self.datasource = None
        else:
            with open(self.info_filename, 'r', encoding='utf8') as f:
                self.info = json.load(f)
            
            self._check_tridesclous_version()
            if len(self.info)>1:
                self._reload_channel_group()
                self._reload_data_source()
                self._reload_data_source_info()
                self._open_processed_data()
            else:
                self.datasource = None
                
            #~ except:
                #~ self.info = {}
                #~ self.flush_info()
                #~ self.datasource = None
 
    def __repr__(self):
        t = "DataIO <id: {}> \n  workdir: {}\n".format(id(self), self.dirname)
        if len(self.info) <= 1 and self.datasource is None:
            t  += "  Not datasource set yet"
            return t
        t += "  sample_rate: {}\n".format(self.sample_rate)
        t += "  total_channel: {}\n".format(self.total_channel)
        if len(self.channel_groups)==1:
            k0, cg0 = next(iter(self.channel_groups.items()))
            ch_names = np.array(self.all_channel_names)[cg0['channels']]
            if len(ch_names)>8:
                chantxt = "[{} ... {}]".format(' '.join(ch_names[:4]),' '.join(ch_names[-4:]))
            else:
                chantxt = "[{}]".format(' '.join(ch_names))
            t += "  channel_groups: {} {}\n".format(k0, chantxt)
        else:
            t += "  channel_groups: {}\n".format(', '.join(['{} ({}ch)'.format(cg, self.nb_channel(cg))
                                                                for cg in self.channel_groups.keys() ]))
        t += "  nb_segment: {}\n".format(self.nb_segment)
        if self.nb_segment<5:
            lengths = [ self.segment_shapes[i][0] for i in range(self.nb_segment)]
            t += '  length: '+' '.join('{}'.format(l) for l in lengths)+'\n'
            t += '  durations: '+' '.join('{:0.1f}'.format(l/self.sample_rate) for l in lengths)+' s.\n'
        if t.endswith('\n'):
            t = t[:-1]
        
        return t

 
    def flush_info(self):
        with open(self.info_filename, 'w', encoding='utf8') as f:
            json.dump(self.info, f, indent=4)
    
    def _check_tridesclous_version(self):
        folder_version= self.info.get('tridesclous_version', 'unknown')
        
        if folder_version == 'unknown':
            w = True
        else:
            v1 = distutils.version.LooseVersion(tridesclous_version).version
            v2 = distutils.version.LooseVersion(self.info['tridesclous_version']).version
            if (v1[0] == v2[0]) and (v1[1] == v2[1]):
                w = False
            else:
                w = True

        if w:
            txt = 'This folder was created with an old tridesclous version ({})\n'\
                    'The actual version is {}\n'\
                    'You may have bug in internal structure.'
            print(txt.format(folder_version, tridesclous_version))
    
    def set_data_source(self, type='RawData', **kargs):
        """
        Set the datasource. Must be done only once otherwise raise error.
        
        Parameters
        ------------------
        
        type: str ('RawData', 'Blackrock', 'Neuralynx', ...)
            The name of the neo.rawio class used to open the dataset.
        kargs: 
            depends on the class used. They are transimted to neo class.
            So see neo doc for kargs
        
        """
        assert type in data_source_classes, 'this source type do not exists yet!!'
        assert 'datasource_type' not in self.info, 'datasource is already set'
        # force abs path name 
        if 'filenames' in kargs:
            kargs['filenames'] = [ os.path.abspath(f) for f in kargs['filenames']]
        elif 'dirnames' in kargs:
            kargs['dirnames'] = [ os.path.abspath(f) for f in kargs['dirnames']]
        else:
            raise ValueError('set_data_source() must contain filenames or dirnames')

        self.info['datasource_type'] = type
        self.info['datasource_kargs'] = kargs
        self._reload_data_source()
        
        # be default chennel group all channels
        channel_groups = {0:{'channels':list(range(self.total_channel))}}
        self.set_channel_groups( channel_groups, probe_filename='default.prb')

        self.flush_info()
        
        self._reload_data_source_info()
        
        # this create segment path
        self._open_processed_data()
    
    def _reload_data_source(self):
        assert 'datasource_type' in self.info
        kargs = self.info['datasource_kargs']
        try:
            self.datasource = data_source_classes[self.info['datasource_type']](**kargs)
        except:
            print('The datatsource is not found', self.info['datasource_kargs'])
            self.datasource = None
        self._reload_data_source_info()
    
    def _save_datasource_info(self):
        assert self.datasource is not None, 'Impossible to load datasource and get info'
        # put some info of datasource
        
        nb_seg = self.datasource.nb_segment
        clean_shape = lambda  shape: tuple(int(e) for e in shape)
        segment_shapes = [clean_shape(self.datasource.get_segment_shape(s)) for s in range(nb_seg)]
        self.info['datasource_info'] = dict(
            total_channel=int(self.datasource.total_channel),
            nb_segment=int(nb_seg),
            sample_rate=float(self.datasource.sample_rate),
            source_dtype=str(self.datasource.dtype),
            all_channel_names=[str(name) for name in self.datasource.get_channel_names()],
            segment_shapes = segment_shapes,
        )
        self.flush_info()

    
    def _reload_data_source_info(self):
        if 'datasource_info' in self.info:
            # no need for datasource
            d = self.info['datasource_info'] 
            self.total_channel = d['total_channel']
            self.nb_segment = d['nb_segment']
            self.sample_rate = d['sample_rate']
            self.source_dtype = np.dtype(d['source_dtype'])
            self.all_channel_names =  d['all_channel_names']
            self.segment_shapes = d['segment_shapes']
        else:
            # This cas is for old directories were
            self._save_datasource_info()
            self._reload_data_source_info()
   
    def _reload_channel_group(self):
        #TODO test in prb is compatible with py3
        d = {}
        probe_filename = os.path.join(self.dirname, self.info['probe_filename'])
        with open(probe_filename) as f:
            exec(f.read(), None, d)
        channel_groups = d['channel_groups']

        for chan_grp, channel_group in channel_groups.items():
            assert 'channels' in channel_group
            channel_group['channels'] = list(channel_group['channels'])
            if 'geometry' not in channel_group or channel_group['geometry'] is None:
                channels = channel_group['channels']
                geometry = self._make_fake_geometry(channels)
                channel_group['geometry'] = geometry
        self.channel_groups = channel_groups
    
    def _rm_old_probe_file(self):
        old_filename = self.info.get('probe_filename', None)
        if old_filename is not None:
            os.remove(os.path.join(self.dirname, old_filename))
        
    def set_probe_file(self, src_probe_filename):
        """
        Set the probe file.
        The probe file is copied inside the working dir.
        
        """
        self._rm_old_probe_file()
        probe_filename = os.path.join(self.dirname, os.path.basename(src_probe_filename))
        try:
            shutil.copyfile(src_probe_filename, probe_filename)
        except shutil.SameFileError:
            # print('probe allready in dir')
            pass
        fix_prb_file_py2(probe_filename)
        # check that the geometry is 2D
        with open(probe_filename) as f:
            d = {}
            exec(f.read(), None, d)
            channel_groups = d['channel_groups']
            for chan_grp, channel_group in channel_groups.items():
                geometry = channel_group.get('geometry', None)
                if geometry is not None:
                    for c, v in geometry.items():
                        assert len(v) == 2, 'Tridesclous need 2D geometry'
                    
                
        
        self.info['probe_filename'] = os.path.basename(probe_filename)
        self.flush_info()
        self._reload_channel_group()
        self._open_processed_data()
        
    
    def download_probe(self, probe_name, origin='kwikteam'):
        """
        Download a prb file from github  into the working dir.
        
        The spiking-circus and kwikteam propose a list prb file.
        See:
           * https://github.com/kwikteam/probes
           * https://github.com/spyking-circus/spyking-circus/tree/master/probes
        
        Parameters
        ------------------
        probe_name: str
            the name of file in github
        origin: 'kwikteam' or 'spyking-circus'
            github project
        
        """
        self._rm_old_probe_file()
        
        
        probe_filename = download_probe(self.dirname, probe_name, origin=origin)
        
        #~ if origin == 'kwikteam':
            #~ #Max Hunter made a list of neuronexus probes, many thanks
            #~ baseurl = 'https://raw.githubusercontent.com/kwikteam/probes/master/'
        #~ elif origin == 'spyking-circus':
            #~ # Pierre Yger made a list of various probe file, many thanks
            #~ baseurl = 'https://raw.githubusercontent.com/spyking-circus/spyking-circus/master/probes/'
        #~ else:
            #~ raise(NotImplementedError)
        
        
        #~ if not probe_name.endswith('.prb'):
            #~ probe_name += '.prb'
        #~ probe_filename = os.path.join(self.dirname,probe_name)
        #~ urlretrieve(baseurl+probe_name, probe_filename)
        #~ fix_prb_file_py2(probe_filename)#fix range to list(range
        fix_prb_file_py2(probe_filename)
        self.info['probe_filename'] = os.path.basename(probe_filename)
        self.flush_info()
        self._reload_channel_group()
        self._open_processed_data()
    
    def _make_fake_geometry(self, channels):
        if len(channels)!=4:
            # assume that it is a linear probes with 100 um
            geometry = { c: [0, i*100] for i, c in enumerate(channels) }
        else:
            # except for tetrode
            geometry = dict(zip(channels, [(0., 50.), (50., 0.), (0., -50.), (-50, 0.)]))
        return geometry
        
    def set_channel_groups(self, channel_groups, probe_filename='channels.prb'):
        """
        Set manually the channel groups dictionary.
        """
        self._rm_old_probe_file()
        
        # checks
        for chan_grp, channel_group in channel_groups.items():
            assert 'channels' in channel_group
            channel_group['channels'] = list(channel_group['channels'])
            if 'geometry' not in channel_group or channel_group['geometry'] is None:
                channels = channel_group['channels']
                #~ geometry = { c: [0, i] for i, c in enumerate(channels) }
                geometry = self._make_fake_geometry(channels)
                channel_group['geometry'] = geometry
        
        # write with hack on json to put key as inteteger (normally not possible in json)
        #~ with open(os.path.join(self.dirname,probe_filename) , 'w', encoding='utf8') as f:
            #~ txt = json.dumps(channel_groups,indent=4)
            #~ for chan_grp in channel_groups.keys():
                #~ txt = txt.replace('"{}":'.format(chan_grp), '{}:'.format(chan_grp))
                #~ for chan in channel_groups[chan_grp]['channels']:
                    #~ txt = txt.replace('"{}":'.format(chan), '{}:'.format(chan))
            #~ txt = 'channel_groups = ' +txt
            #~ f.write(txt)
        
        create_prb_file_from_dict(channel_groups, os.path.join(self.dirname,probe_filename))
        
        
        
        self.info['probe_filename'] = probe_filename
        self.flush_info()
        self._reload_channel_group()
        self._open_processed_data()
    
    def add_one_channel_group(self, channels=[], chan_grp=0, geometry=None):
        channels = list(channels)
        if geometry is None:
            geometry = self._make_fake_geometry(channels)
        
        self.channel_groups[chan_grp] = {'channels': channels, 'geometry':geometry}
        #rewrite with same name
        self.set_channel_groups(self.channel_groups, probe_filename=self.info['probe_filename'])
        
    def get_geometry(self, chan_grp=0):
        """
        Get the geometry for a given channel group in a numpy array way.
        """
        channel_group = self.channel_groups[chan_grp]
        geometry = [ channel_group['geometry'][chan] for chan in channel_group['channels'] ]
        geometry = np.array(geometry, dtype='float64')
        return geometry
    
    def get_channel_distances(self, chan_grp=0):
        geometry = self.get_geometry(chan_grp=chan_grp)
        distances = sklearn.metrics.pairwise.euclidean_distances(geometry)
        return distances
    
    def get_channel_adjacency(self, chan_grp=0, adjacency_radius_um=None):
        assert adjacency_radius_um is not None
        channel_distances = self.get_channel_distances(chan_grp=chan_grp)
        channels_adjacency = {}
        nb_chan = self.nb_channel(chan_grp=chan_grp)
        for c in range(nb_chan):
            nearest, = np.nonzero(channel_distances[c, :] < adjacency_radius_um)
            channels_adjacency[c] = nearest
        return channels_adjacency

    def nb_channel(self, chan_grp=0):
        """
        Number of channel for a channel group.
        """
        #~ print('DataIO.nb_channel', self.channel_groups)
        return len(self.channel_groups[chan_grp]['channels'])


    def channel_group_label(self, chan_grp=0):
        """
        Label of channel for a group.
        """
        label = 'chan_grp {} - '.format(chan_grp)
        
        channels = self.channel_groups[chan_grp]['channels']
        ch_names = np.array(self.all_channel_names)[channels]
        
        if len(ch_names)<8:
            label += ' '.join(ch_names)
        else:
            label += ' '.join(ch_names[:3]) + ' ... ' + ' '.join(ch_names[-2:])
        
        return label


    
    def _open_processed_data(self):
        self.channel_group_path = {}
        self.segments_path = {}
        for chan_grp in self.channel_groups.keys():
            self.segments_path[chan_grp] = []
            cg_path = os.path.join(self.dirname, 'channel_group_{}'.format(chan_grp))
            self.channel_group_path[chan_grp] = cg_path
            if not os.path.exists(cg_path):
                os.mkdir(cg_path)
            for i in range(self.nb_segment):
                segment_path = os.path.join(cg_path, 'segment_{}'.format(i))
                if not os.path.exists(segment_path):
                    os.mkdir(segment_path)
                self.segments_path[chan_grp].append(segment_path)
        
        self.arrays = {}
        for chan_grp in self.channel_groups.keys():
            self.arrays[chan_grp] = []
            
            for i in range(self.nb_segment):
                arrays = ArrayCollection(parent=None, dirname=self.segments_path[chan_grp][i])
                self.arrays[chan_grp].append(arrays)
            
                for name in ['processed_signals', 'spikes']:
                    self.arrays[chan_grp][i].load_if_exists(name)
    
    def get_segment_length(self, seg_num):
        """
        Segment length (in sample) for a given segment index
        """
        full_shape =  self.segment_shapes[seg_num]
        return full_shape[0]
    
    def get_segment_shape(self, seg_num, chan_grp=0):
        """
        Segment shape for a given segment index and channel group.
        """
        full_shape =  self.segment_shapes[seg_num]
        shape = (full_shape[0], self.nb_channel(chan_grp))
        return shape
    
    def get_duration_per_segments(self, total_duration=None):
        duration_per_segment = []
        
        if total_duration is not None:
            remain = float(total_duration)
        
        for seg_num in range(self.nb_segment):
            dur = self.get_segment_length(seg_num=seg_num) / self.sample_rate
            
            if total_duration is None:
                duration_per_segment.append(dur)
            elif remain ==0:
                duration_per_segment.append(0.)
            elif dur <=remain:
                duration_per_segment.append(dur)
                remain -= dur
            else:
                duration_per_segment.append(remain)
                remain = 0.
        
        return duration_per_segment
    
    def get_signals_chunk(self, seg_num=0, chan_grp=0,
                i_start=None, i_stop=None,
                signal_type='initial', pad_width=0):
        """
        Get a chunk of signal for for a given segment index and channel group.
        
        The signal can be the 'initial' (aka raw signal), the none filetered signals or
        the 'processed' signal.
        
        Parameters
        ------------------
        seg_num: int
            segment index
        chan_grp: int
            channel group key
        i_start: int or None
           start index (included)
        i_stop: int or None
            stop index (not included)
        signal_type: str
            'initial' or 'processed'
        pad_width: int (0 default)
            Add optional pad on each sides
            usefull for filtering border effect
        
        """
        channels = self.channel_groups[chan_grp]['channels']
        
        after_padding = False
        after_padding_left = 0
        after_padding_right = 0
        if pad_width > 0:
            i_start = i_start - pad_width
            i_stop =  i_stop + pad_width
        
            if i_start < 0:
                after_padding = True
                after_padding_left = -i_start
                i_start = 0
            if i_stop > self.get_segment_length(seg_num):
                after_padding = True
                after_padding_right = i_stop - self.get_segment_length(seg_num)
                i_stop = self.get_segment_length(seg_num)
        
        if signal_type=='initial':
            data = self.datasource.get_signals_chunk(seg_num=seg_num, i_start=i_start, i_stop=i_stop)
            data = data[:, channels]
        elif signal_type=='processed':
            data = self.arrays[chan_grp][seg_num].get('processed_signals')[i_start:i_stop, :]
        else:
            raise(ValueError, 'signal_type is not valide')
        
        if after_padding:
            # finalize padding on border
            data2 = np.zeros((data.shape[0] + after_padding_left + after_padding_right, data.shape[1]), dtype=data.dtype)
            data2[after_padding_left:data2.shape[0]-after_padding_right, :] = data
            data = data2
        
        return data
    
    def iter_over_chunk(self, seg_num=0, chan_grp=0,  i_stop=None,
                        chunksize=1024, pad_width=0, with_last_chunk=False,   **kargs):
        """
        Create an iterable on signals. ('initial' or 'processed')
        
        Usage
        ----------
        
            for ind, sig_chunk in data.iter_over_chunk(seg_num=0, chan_grp=0, chunksize=1024, signal_type='processed'):
                do_something_on_chunk(sig_chunk)
        
        """
        seg_length = self.get_segment_length(seg_num)
        
        length = seg_length
        if i_stop is not None:
            length = min(length, i_stop)
        
        total_length = length + pad_width
        
        nloop = total_length//chunksize
        if total_length % chunksize and with_last_chunk:
            nloop += 1
        
        last_sample = None
        for i in range(nloop):
            i_stop = (i+1)*chunksize
            i_start = i_stop - chunksize
            
            if i_stop > seg_length:
                sigs_chunk2 = np.zeros((chunksize, sigs_chunk.shape[1]), dtype=sigs_chunk.dtype)
                if i_start < seg_length:
                    sigs_chunk = self.get_signals_chunk(seg_num=seg_num, chan_grp=chan_grp, i_start=i_start, i_stop=seg_length, **kargs)
                    sigs_chunk2[:sigs_chunk.shape[0], :] = sigs_chunk
                    last_sample = sigs_chunk[-1, :]
                    # extend with last sample : agttenuate fileter border effect
                    sigs_chunk2[sigs_chunk.shape[0]:, :] = last_sample
                else:
                    if last_sample is not None:
                        sigs_chunk2[:, :] = last_sample
                yield  i_stop, sigs_chunk2
                
            else:
                sigs_chunk = self.get_signals_chunk(seg_num=seg_num, chan_grp=chan_grp, i_start=i_start, i_stop=i_stop, **kargs)
                if i_stop == seg_length:
                    last_sample = sigs_chunk[-1, :]
                yield  i_stop, sigs_chunk
        
        #~ if with_last_chunk and i_stop<total_length:
            #~ i_start = i_stop
            #~ i_stop = length
            #~ sigs_chunk = self.get_signals_chunk(seg_num=seg_num, chan_grp=chan_grp, i_start=i_start, i_stop=i_stop, **kargs)
            
            #~ sigs_chunk2 = np.zeros((chunksize, sigs_chunk.shape[1]), dtype=sigs_chunk.dtype)
            #~ if sigs_chunk.shape[0] > 0:
                #~ sigs_chunk2[:sigs_chunk.shape[0], :] = sigs_chunk
                #~ # extend with last sample : agttenuate fileter border effect
                #~ sigs_chunk2[sigs_chunk.shape[0]:, :] = sigs_chunk[-1, :]
            
            #~ yield  i_start+chunksize, sigs_chunk2
    
    def reset_processed_signals(self, seg_num=0, chan_grp=0, dtype='float32'):
        """
        Reset processed signals.
        """
        self.arrays[chan_grp][seg_num].create_array('processed_signals', dtype, 
                            self.get_segment_shape(seg_num, chan_grp=chan_grp), 'memmap')
        self.arrays[chan_grp][seg_num].annotate('processed_signals', processed_length=0)
    
    def set_signals_chunk(self,sigs_chunk, seg_num=0, chan_grp=0, i_start=None, i_stop=None, signal_type='processed'):
        """
        Set a signal chunk (only for 'processed')
        """
        assert signal_type != 'initial'

        if signal_type=='processed':
            data = self.arrays[chan_grp][seg_num].get('processed_signals')
            data[i_start:i_stop, :] = sigs_chunk
        
    def flush_processed_signals(self, seg_num=0, chan_grp=0, processed_length=-1):
        """
        Flush the underlying memmap for processed signals.
        """
        self.arrays[chan_grp][seg_num].flush_array('processed_signals')
        self.arrays[chan_grp][seg_num].annotate('processed_signals', processed_length=processed_length)
    
    def get_processed_length(self, seg_num=0, chan_grp=0):
        """
        Get the length in sample how already processed part of the segment.
        """
        return self.arrays[chan_grp][seg_num].get_annotation('processed_signals', 'processed_length')
    
    def already_processed(self, seg_num=0, chan_grp=0, length=None):
        """
        Check if the segment is entirely processedis already computed until length
        """
        # check if signals are processed
        if length is None:
            length = self.get_segment_length(seg_num)
        already_done = self.get_processed_length(seg_num, chan_grp=chan_grp)
        return  already_done >= length
    
    def reset_spikes(self, seg_num=0,  chan_grp=0, dtype=None):
        """
        Reset spikes.
        """
        assert dtype is not None
        self.arrays[chan_grp][seg_num].initialize_array('spikes', 'memmap', dtype, (-1,))
        
    def append_spikes(self, seg_num=0, chan_grp=0, spikes=None):
        """
        Append spikes.
        """
        if spikes is None: return
        self.arrays[chan_grp][seg_num].append_chunk('spikes', spikes)
        
    def flush_spikes(self, seg_num=0, chan_grp=0):
        """
        Flush underlying memmap for spikes.
        """
        self.arrays[chan_grp][seg_num].finalize_array('spikes')
    
    def is_spike_computed(self, chan_grp=0):
        done = all(self.arrays[chan_grp][seg_num].has_key('spikes') for seg_num in range(self.nb_segment))
        return done
    
    def get_spikes(self, seg_num=0, chan_grp=0, i_start=None, i_stop=None):
        """
        Read spikes
        """
        if not self.arrays[chan_grp][seg_num].has_key('spikes'):
            return None
        spikes = self.arrays[chan_grp][seg_num].get('spikes')
        if spikes is None:
            return
        return spikes[i_start:i_stop]

    def get_peak_values(self,  seg_num=0, chan_grp=0, sample_indexes=None, channel_indexes=None):
        """
        Extract peak values
        """
        assert sample_indexes is not None, 'Provide sample_indexes'
        assert channel_indexes is not None, 'Provide channel_indexes'
        sigs = self.arrays[chan_grp][seg_num].get('processed_signals')
        
        peak_values = []
        for s, c in zip(sample_indexes, channel_indexes):
            peak_values.append(sigs[s, c])
        peak_values = np.array(peak_values)
        return peak_values
        
    
    def get_some_waveforms(self, seg_num=None, seg_nums=None, chan_grp=0, peak_sample_indexes=None,
                                n_left=None, n_right=None, waveforms=None, channel_indexes=None):
        """
        Exctract some waveforms given sample_indexes
        
        seg_num is int then all spikes come from same segment
        
        if seg_num is None then seg_nums is an array that contain seg_num for each spike.
        """
        assert peak_sample_indexes is not None, 'Provide sample_indexes'
        peak_width = n_right - n_left
        
        if channel_indexes is None:
            nb_chan = self.nb_channel(chan_grp)
        else:
            nb_chan = len(channel_indexes)
        
        if waveforms is None:
            dtype = self.arrays[chan_grp][0].get('processed_signals').dtype
            waveforms = np.zeros((peak_sample_indexes.size, peak_width, nb_chan), dtype=dtype)
        else:
            assert waveforms.shape[0] == peak_sample_indexes.size
            assert waveforms.shape[1] == peak_width
        
        if seg_num is not None:
            left_indexes = peak_sample_indexes + n_left
            sigs = self.arrays[chan_grp][seg_num].get('processed_signals')
            extract_chunks(sigs, left_indexes, peak_width, 
                                        channel_indexes=channel_indexes, chunks=waveforms)
        elif seg_nums is not None and isinstance(seg_nums, np.ndarray):
            n = 0
            for seg_num in np.unique(seg_nums):
                
                mask = seg_num == seg_nums
                left_indexes = peak_sample_indexes[mask]+n_left
                if left_indexes.size == 0:
                    continue
                chunks = waveforms[n:n+left_indexes.size] # this avoid a copy
                sigs = self.arrays[chan_grp][seg_num].get('processed_signals')
                extract_chunks(sigs, left_indexes, peak_width, 
                                            channel_indexes=channel_indexes, chunks=chunks)
                n += left_indexes.size
        else:
            raise 'error seg_num or seg_nums'
        
        return waveforms


    def save_catalogue(self, catalogue, name='initial'):
        """
        Save the catalogue made by `CatalogueConstructor` and needed
        by `Peeler` inside the working dir.
        
        Note that you can construct several catalogue for the same dataset
        to compare then just change the name. Different folder name so.
        
        """
        catalogue = dict(catalogue)
        chan_grp = catalogue['chan_grp']
        dir = os.path.join(self.dirname,'channel_group_{}'.format(chan_grp), 'catalogues', name)
        if not os.path.exists(dir):
            os.makedirs(dir)
        arrays = ArrayCollection(parent=None, dirname=dir)
        
        to_rem = []
        for k, v in catalogue.items():
            if isinstance(v, np.ndarray):
                arrays.add_array(k, v, 'memmap')
                to_rem.append(k)
        
        for k in to_rem:
            catalogue.pop(k)
        
        # JSON is not possible for now because some key in catalogue are integer....
        # So bad....
        #~ with open(os.path.join(dir, 'catalogue.json'), 'w', encoding='utf8') as f:
            #~ json.dump(catalogue, f, indent=4)
        with open(os.path.join(dir, 'catalogue.pickle'), 'wb') as f:
            pickle.dump(catalogue, f)
        
    
    def load_catalogue(self,  name='initial', chan_grp=0):
        """
        Load the catalogue dict.
        """
        dir = os.path.join(self.dirname,'channel_group_{}'.format(chan_grp), 'catalogues', name)
        filename = os.path.join(dir, 'catalogue.pickle')
        #~ with open(os.path.join(dir, 'catalogue.json'), 'r', encoding='utf8') as f:
                #~ catalogue = json.load(f)
        if not os.path.exists(filename):
            return
        
        with open(filename, 'rb') as f:
            catalogue = pickle.load(f)

        
        arrays = ArrayCollection(parent=None, dirname=dir)
        arrays.load_all()
        for k in arrays.keys():
            catalogue[k] = np.array(arrays.get(k), copy=True)
        
        
        return catalogue
    
    def export_spikes(self, export_path=None,
                split_by_cluster=False,  use_cell_label=True, formats=None):
        """
        Export spikes to other format (csv, matlab, excel, ...)
        
        Parameters
        ------------------
        export_path: str or None
           export path. If None (default then inside working dir)
        split_by_cluster: bool (default False)
           Each cluster is split to a diffrent file or not.
        use_cell_label: bool (deafult True)
            if true cell_label is used if false cluster_label is used
        formats: 'csv' or 'mat' or 'xlsx'
            The output format.
        """
        
        if export_path is None:
            export_path = os.path.join(self.dirname, 'export')
        
        if formats is None:
            exporters = export_list
        elif isinstance(formats, str):
            assert formats in export_dict
            exporters = [ export_dict[formats] ]
        elif isinstance(format, list):
            exporters = [ export_dict[format] for format in formats]
        else:
            return
        
        for chan_grp in self.channel_groups.keys():
            
            catalogue = self.load_catalogue(chan_grp=chan_grp)
            if catalogue is None:
                continue
            
            if not self.is_spike_computed(chan_grp=chan_grp):
                continue
            
            for seg_num in range(self.nb_segment):
                spikes = self.get_spikes(seg_num=seg_num, chan_grp=chan_grp)
                
                if spikes is None: continue
                
                args = (spikes, catalogue, seg_num, chan_grp, export_path,)
                kargs = dict(split_by_cluster=split_by_cluster, use_cell_label=use_cell_label)
                for exporter in exporters:
                    exporter(*args, **kargs)
    
    def get_log_path(self, chan_grp=0):
        cg_path = os.path.join(self.dirname, 'channel_group_{}'.format(chan_grp))
        log_path = os.path.join(cg_path, 'log')
        if not os.path.exists(log_path):
            os.makedirs(log_path)
        
        return log_path



"""
This is automatic home made method for clustering in TDC.

It is based on hdbscan algo.


"""


import numpy as np
import os
import time

import sklearn
import sklearn.cluster
import sklearn.mixture
import sklearn.metrics






from .dip import diptest


import hdbscan

try:
    import isosplit5
    HAVE_ISOSPLIT5 = True
except:
    HAVE_ISOSPLIT5 = False


class PruningShears:
    def __init__(self,  
                        features,
                        channel_to_features,
                        peaks,
                        peak_index, # in catalogue constructor size
                        noise_features,
                        n_left, n_right,
                        peak_sign, threshold,
                        geometry,
                        dense_mode,
                        catalogueconstructor, # for get_some_waveform TODO find better
                        
                        
                        adjacency_radius_um=50,
                        high_adjacency_radius_um = 30,
                        
                        min_cluster_size=20,
                        max_loop=1000,
                        break_nb_remain=30,
                        auto_merge_threshold=2.,
                        max_per_cluster_for_median = 300,
                        n_components_local_pca=3,
                        
                        print_debug=False,
                        debug_plot=False
                        ):
        
   
        self.features = features
        self.channel_to_features = channel_to_features
        self.peaks = peaks
        self.peak_index = peak_index
        self.noise_features = noise_features
        self.n_left = n_left
        self.n_right = n_right
        self.width = n_right - n_left
        self.peak_sign = peak_sign
        self.threshold = threshold
        
        self.geometry = geometry
        self.dense_mode =  dense_mode
        self.catalogueconstructor = catalogueconstructor
        self.cc = catalogueconstructor
        
        
        
        #user params
        self.adjacency_radius_um = adjacency_radius_um
        self.high_adjacency_radius_um = high_adjacency_radius_um
        
        self.min_cluster_size = min_cluster_size
        self.max_loop = max_loop
        self.break_nb_remain = break_nb_remain
        self.auto_merge_threshold = auto_merge_threshold
        self.print_debug = print_debug
        self.max_per_cluster_for_median = max_per_cluster_for_median
        self.n_components_local_pca = n_components_local_pca
        
        self.nb_channel = self.channel_to_features.shape[0]
        
        self.channel_distances = sklearn.metrics.pairwise.euclidean_distances(geometry)
        #~ print('self.channel_distances', self.channel_distances)
        
        if self.dense_mode:
            self.channel_adjacency = {c: np.arange(self.nb_channel) for c in range(self.nb_channel)}
            self.channel_high_adjacency = {c: np.arange(self.nb_channel) for c in range(self.nb_channel)}
            
        else:
            self.channel_adjacency = {}
            for c in range(self.nb_channel):
                nearest, = np.nonzero(self.channel_distances[c, :] < self.adjacency_radius_um)
                self.channel_adjacency[c] = nearest
            
            self.channel_high_adjacency = {}
            for c in range(self.nb_channel):
                nearest, = np.nonzero(self.channel_distances[c, :] < self.high_adjacency_radius_um)
                self.channel_high_adjacency[c] = nearest
        
        #~ print('self.channel_adjacency', self.channel_adjacency)
        #~ print('adjacency_radius_um', adjacency_radius_um)
        


        
        #~ self.debug_plot = False
        #~ self.debug_plot = True
        self.debug_plot = debug_plot
        
        

    def log(self, *args, **kargs):
        if self.print_debug:
            print(*args, **kargs)
    
    def do_the_job(self):
        self.centroids = {}
        
        t0 = time.perf_counter()
        cluster_labels = self.explore_split_loop()
        t1 = time.perf_counter()
        self.log('explore_split_loop', t1-t0)
        

        
        return cluster_labels
    
    def next_channel(self, mask_loop, chan_visited):
        self.log('next_channel percentiles', 'peak_max.size', np.sum(mask_loop))
        self.log('chan_visited', chan_visited)
        
        #~ peak_max = self.all_peak_max[mask_loop, :]
        
        #~ self.peaks['channel'][mask_loop]
        
        
        percentiles = np.zeros(self.nb_channel)
        for c in range(self.nb_channel):
            if c in chan_visited:
                continue
            
            #~ x = peak_max[:, c]
            #~ x = x[x>self.threshold]
            
            #~ if x.size>self.min_cluster_size * 4:
                #~ per = np.nanpercentile(x, 99)
                #~ per = np.nanpercentile(x, 95)
                #~ per = np.nanpercentile(x, 90)
                #~ per = np.nanpercentile(x, 80)
            
            #~ elif x.size>self.min_cluster_size:
                #~ per = np.nanpercentile(x, 60)
            #~ else:
                #~ per = 0
            #~ percentiles[c] = per
            
            mask = self.peaks['channel'][mask_loop] == c
            #~ print(mask.size, peak_max.shape)
            #~ x = peak_max[mask, :][:, c]
            x = np.abs(self.peaks['extremum_amplitude'][mask_loop][mask])
            if x.size >self.min_cluster_size:
                per = np.nanpercentile(x, 90)
            else:
                per = 0
            percentiles[c] = per
            
            
        
        #~ mask = percentiles > 0
        #~ print('mask.size', mask.size, 'mask.sum', mask.sum())
        
        order_visit = np.argsort(percentiles)[::-1]
        percentiles = percentiles[order_visit]
        
        mask = (percentiles > 0) & ~np.in1d(order_visit, chan_visited)
        
        order_visit = order_visit[mask]
        percentiles = percentiles[mask]
        
        #~ print('percentiles', percentiles)
        self.log('order_visit :10', order_visit[:10])
        self.log('percentiles :10', percentiles[:10])
        
        if len(order_visit)==0:
            self.log('len(order_visit)==0')
            return None
        
        actual_chan = order_visit[0]
        #~ if len(order_visit)>3:
            #~ # add some randomness
            #~ order_visit = np.random.permutation(order_visit[:order_visit.size//3])
            #~ actual_chan = order_visit[0]
        #~ else:
            #~ actual_chan = order_visit[0]
        
        return actual_chan
    
    def one_sub_cluster(self, local_data, allow_single_cluster):
        # clustering on one channel or one adjacency
        
        
        #~ n_components = min(local_data.shape[1], self.n_components_local_pca)
        #~ pca =  sklearn.decomposition.IncrementalPCA(n_components=n_components, whiten=True)
        
        n_components = min(local_data.shape[1]-1, self.n_components_local_pca)
        pca =  sklearn.decomposition.TruncatedSVD(n_components=n_components)
        
        local_features = pca.fit_transform(local_data)
        
        
        clusterer = hdbscan.HDBSCAN(min_cluster_size=self.min_cluster_size, allow_single_cluster=allow_single_cluster, metric='l2')
        #~ clusterer = hdbscan.HDBSCAN(min_cluster_size=self.min_cluster_size, allow_single_cluster=True)
        #~ clusterer = hdbscan.HDBSCAN(min_cluster_size=100, min_samples=20, allow_single_cluster=True)
        
        #~ t0 = time.perf_counter()
        #~ local_labels = clusterer.fit_predict(local_data)
        #~ t1 = time.perf_counter()
        #~ print('fit_predict wf', t1-t0)

        
        #~ t0 = time.perf_counter()
        local_labels = clusterer.fit_predict(local_features)
        #~ t1 = time.perf_counter()
        #~ print('fit_predict pca', t1-t0)
        
        
        
        # try isosplit here not stable enought on windows
        #~ local_labels = isosplit5.isosplit5(local_data.T)
        
        return local_labels
    
    def check_candidate_labels(self, local_ind, local_labels, actual_chan):
        # keep only cluster when:
        #   * template peak is center on self.n_left
        #   * cluster size is bug enoutgh
        
        
        
        #~ candidate_labels = []
        peak_is_on_chan = []
        peak_is_aligned = []
        best_chan_peak_values = []
        best_chan = []
        cluster_sizes = []
        #~ actual_chan_peak_values =  []
        
        
        local_centroids = {}
        
        unique_labels = np.unique(local_labels)
        unique_labels = unique_labels[unique_labels>=0]
        
        for l, label in enumerate(unique_labels):


            ind = local_ind[label == local_labels]
            cluster_size = ind.size
            cluster_sizes.append(cluster_size)
            if ind.size > self.max_per_cluster_for_median:
                sub_sel = np.random.choice(ind.size, self.max_per_cluster_for_median, replace=False)
                ind = ind[sub_sel]
            
            if self.dense_mode:
                #~ centroid = np.median(self.waveforms[ind, :, :], axis=0)
                wfs = self.cc.get_some_waveforms( peaks_index=self.peak_index[ind], channel_indexes=None)
                centroid = np.median(wfs, axis=0)
            else:
                #~ centroid = np.median(self.waveforms[ind, :, :][:, :, local_channels], axis=0)
                local_channels = self.channel_adjacency[actual_chan]
                wfs = self.cc.get_some_waveforms( peaks_index=self.peak_index[ind], channel_indexes=local_channels)
                centroid = np.median(wfs, axis=0)
            
            local_centroids[label] = centroid
            
            if self.peak_sign == '-':
                chan_peak_local = np.argmin(np.min(centroid, axis=0))
                pos_peak = np.argmin(centroid[:, chan_peak_local])
            elif self.peak_sign == '+':
                chan_peak_local = np.argmax(np.max(centroid, axis=0))
                pos_peak = np.argmax(centroid[:, chan_peak_local])
            
            if self.dense_mode:
                # for dense we don't care the channel
                # and alignement
                chan_peak = chan_peak_local
                peak_is_on_chan.append(True)
                aligned = True
            else:
                chan_peak = local_channels[chan_peak_local]
                peak_is_on_chan.append(chan_peak == actual_chan)
                best_chan.append(chan_peak)
                aligned = np.abs(-self.n_left - pos_peak) <= 1
                
            self.log('-self.n_left', -self.n_left, 'pos_peak', pos_peak, 'aligned', aligned)
            peak_is_aligned.append(aligned)
            
            best_chan_peak_value = np.abs(centroid[pos_peak, chan_peak_local])
            best_chan_peak_values.append(best_chan_peak_value)
            
            self.log('label', label, 'chan_peak', chan_peak, 'chan_peak_local', chan_peak_local, 'aligned', aligned, 'best_chan_peak_value', best_chan_peak_value, 'cluster_size', cluster_size, 'peak_is_on_chan', peak_is_on_chan[-1])
            #~ if peak_is_aligned and cluster_size>self.min_cluster_size:
                #~ candidate_labels.append(label)
        
        cluster_sizes = np.array(cluster_sizes)
        peak_is_aligned = np.array(peak_is_aligned)
        peak_is_on_chan = np.array(peak_is_on_chan)
        best_chan_peak_values = np.array(best_chan_peak_values)
        best_chan = np.array(best_chan)
        
        if unique_labels.size>0:
            #~ print(cluster_sizes.shape, peak_is_aligned.shape, peak_is_on_chan.shape)
            candidate_mask= (cluster_sizes>=self.min_cluster_size) & peak_is_aligned & peak_is_on_chan
            #~ candidate_labels = unique_labels[candidate_mask]
            
            elsewhere_mask= (cluster_sizes>=self.min_cluster_size) & peak_is_aligned & ~peak_is_on_chan
            #~ candidate_labels_elsewhere = unique_labels[elsewhere_mask]
            
        else:
            candidate_mask = np.array([], dtype='bool')
            elsewhere_mask = np.array([], dtype='bool')
        
        
        return unique_labels, candidate_mask, elsewhere_mask, best_chan_peak_values,\
                        best_chan, peak_is_aligned, peak_is_on_chan, local_centroids

    
    def explore_split_loop(self):
        #~ print('all_peak_max')
        #~ ind_peak = -self.n_left
        #~ self.all_peak_max = self.waveforms[:, ind_peak, : ].copy()
        #~ if self.peak_sign == '-' :
            #~ self.all_peak_max *= -1
        
        #~ print('all_peak_max done')
        
        #~ nb_channel = self.waveforms.shape[2]
        
        if self.dense_mode:
            # global pca so no sparse used
            pass
            
        else:
            # general case 
            n_components_by_channel = self.features.shape[1] // self.nb_channel
            
            #~ self.weights_feat_chans = {}
            self.mask_feat_per_chan = {}
            self.mask_feat_per_adj = {}
            for chan in range(self.nb_channel):
                adjacency = self.channel_adjacency[chan]
                
                # mask feature per channel
                mask_feat = np.zeros(self.features.shape[1], dtype='bool')
                mask_feat[chan*n_components_by_channel:(chan+1)*n_components_by_channel] = True
                self.mask_feat_per_chan[chan] = mask_feat
                
                # mask_feat per channel adjacency
                mask_feat = np.zeros(self.features.shape[1], dtype='bool')
                for i in range(n_components_by_channel):
                    mask_feat[adjacency*n_components_by_channel+i] = True
                self.mask_feat_per_adj[chan] = mask_feat
                
                # weights
                #~ weights = []
                #~ for adj_chan in adjacency:
                    #~ d = self.channel_distances[chan, adj_chan]
                    #~ w = np.exp(-d/self.adjacency_radius_um * 2) # TODO fix this factor 2
                    #~ print('chan', chan, 'adj_chan', adj_chan, 'd', d, 'w', w)
                    #~ weights += [ w ] * n_components_by_channel
                #~ self.weights_feat_chans[chan] = np.array(weights).reshape(1, -1)
        
        force_next_chan = None
        
        cluster_labels = np.zeros(self.features.shape[0], dtype='int64')
        last_nb_remain = cluster_labels.size
        
        k = 0
        chan_visited = []
        #~ for iloop in range(self.max_loop):
        iloop = -1
        while True:
            iloop += 1
            
            nb_remain = np.sum(cluster_labels>=k)
            mask_loop = cluster_labels == k
            nb_working = np.sum(mask_loop)
            self.log()
            self.log('iloop', iloop, 'k', k, 'nb_remain', nb_remain, 'nb_working', nb_working, {True:'full', False:'partial'}[nb_remain==nb_working])
            
            if iloop>=self.max_loop:
                cluster_labels[mask_loop] = -1
                self.log('BREAK iloop', iloop)
                print('Warning PruningShears reach max_loop limit there are maybe more cluster')
                break
            
            if iloop!=0 and nb_remain<self.break_nb_remain:
                cluster_labels[mask_loop] = -1
                self.log('BREAK nb_remain', nb_remain, '<', self.break_nb_remain)
                break
            
            if nb_working<self.min_cluster_size:
                self.log('TRASH: too few. Explore remain')
                cluster_labels[mask_loop] = -1
                k += 1
                chan_visited = []
                force_next_chan = None
                continue
            
            if self.dense_mode:
                actual_chan = 0
            else:
                if force_next_chan is None:
                    actual_chan = self.next_channel(mask_loop,  chan_visited)
                elif force_next_chan is not None and force_next_chan in chan_visited:
                    # this is impossible normally
                    self.log('!!!!! impossible force', force_next_chan, 'chan_visited', chan_visited)
                    raise # TODO
                    #~ actual_chan = self.next_channel(peak_max, chan_visited)
                else:
                    actual_chan = force_next_chan
                    self.log('force', actual_chan)

            
            if actual_chan is None:
                # no more chan to explore
                #~ cluster_labels[mask_loop] = -1
                #~ self.log('BREAK actual_chan None')
                #~ self.log('BREAK actual_chan None')
                #~ break
                
                if self.dense_mode:
                    pass
                    # impossible because actual_chan always 0
                else:
                    # start again explore all channels
                    if nb_remain < last_nb_remain:
                        cluster_labels[mask_loop] += 1
                        k += 1
                        chan_visited = []
                        force_next_chan = None
                        last_nb_remain = nb_remain
                        self.log('Explore again all channel once')
                        continue
                    else:
                        self.log('Stop exploring again all channel once')
                        break
                    
            
            self.log('actual_chan', actual_chan)
            
            if not self.dense_mode:
                adjacency = self.channel_adjacency[actual_chan]
                high_adjacency = self.channel_high_adjacency[actual_chan]
                self.log('adjacency', adjacency, 'high_adjacency', high_adjacency)
            
            
            mask_thresh = np.zeros(mask_loop.size, dtype='bool')
            if self.dense_mode:
                mask_thresh[:] = True
            else:
                #~ peak_max = self.all_peak_max[mask_loop, :]
                #~ mask_thresh[mask_loop] = peak_max[:, actual_chan] > self.threshold
                mask_thresh[mask_loop] = np.in1d(self.peaks['channel'][mask_loop], high_adjacency)
                
            self.log('mask_loop.size', mask_loop.size, 'mask_loop.sum', mask_loop.sum(), 'mask_thresh.sum', mask_thresh.sum())
            
            ind_l0,  = np.nonzero(mask_loop & mask_thresh)
            self.log('ind_l0.size', ind_l0.size)
            
            if ind_l0.size < self.min_cluster_size:
                force_next_chan = None
                chan_visited.append(actual_chan)
                self.log('len(ind_l0.size) < self.min_cluster_size')
                features_l0 = None
                labels_l0 = None
                possible_labels_l0 = None
                candidate_labels_l0 = None
                final_label = None
                self._plot_debug(actual_chan, ind_l0, features_l0, labels_l0, possible_labels_l0, candidate_labels_l0, final_label)
                continue
            
            
            # Step2: 
            if self.dense_mode:
                wf_features = self.features.take(ind_l0, axis=0)
            else:
                mask_feat = self.mask_feat_per_adj[actual_chan]
                wf_features = self.features.take(ind_l0, axis=0).compress(mask_feat, axis=1)
            
            self.log('wf_features.shape', wf_features.shape)
            
            features_l0 = wf_features
            self.log('features_l0.shape', features_l0.shape)
            
            labels_l0 = self.one_sub_cluster(features_l0, allow_single_cluster=False)
            
            possible_labels_l0 = np.unique(labels_l0)
            possible_labels_l0 = possible_labels_l0[possible_labels_l0>=0]
            
            if len(possible_labels_l0) == 0:
                # add noise to force one cluster to be discovered
                if self.dense_mode:
                    noise_features = self.noise_features
                else:
                    noise_features = self.noise_features.compress(mask_feat, axis=1)
                self.log('Add noise to force one cluster')
                features_with_noise = np.concatenate([noise_features, wf_features, ], axis=0)
                labels_l0 = self.one_sub_cluster(features_with_noise, allow_single_cluster=False)
                labels_l0 = labels_l0[self.noise_features.shape[0]:]
                possible_labels_l0 = np.unique(labels_l0)
                possible_labels_l0 = possible_labels_l0[possible_labels_l0>=0]
                
                # instead of experimental save trash
                if len(possible_labels_l0) == 0:
                    if self.dense_mode:
                        self.log('no more with dense mode')
                        break
                    else:
                
                        self.log('Add noise do not work, explore other channel')
                        force_next_chan = None
                        chan_visited.append(actual_chan)
                        final_label = None
                        candidate_labels_l0 = None
                        final_label = None
                        self._plot_debug(actual_chan, ind_l0, features_l0, labels_l0, possible_labels_l0, candidate_labels_l0, final_label)
                        continue
            
            #~ if len(possible_labels_l0) == 0:
                #~ # experimental!!!!!!
                #~ pca =  sklearn.decomposition.IncrementalPCA(n_components=2, whiten=True)
                #~ reduced_features = pca.fit_transform(wf_features)
                
                #~ pval = diptest(np.sort(reduced_features[:, 0]), numt=200)
                
                #~ if pval is not None and pval>0.2:
                    #~ self.log('len(possible_labels_l0) == 0 BUT diptest pval', pval)

                    #~ if self.dense_mode:
                        #~ wf_l0 = self.waveforms.take(ind_l0, axis=0)
                    #~ else:
                        #~ mask_feat = self.mask_feat_per_adj[actual_chan]
                        #~ wf_l0 = self.waveforms.take(ind_l0, axis=0).take(adjacency, axis=2)
                    #~ centroid = np.median(wf_l0, axis=0)
                    #~ centroid = centroid[np.newaxis, :, :]
                    #~ out_up = np.any(np.any(wf_l0 > centroid + 4, axis=2), axis=1)
                    #~ out_dw = np.any(np.any(wf_l0 < centroid - 4, axis=2), axis=1)
                    #~ ok = ~out_up & ~out_dw
                    #~ if np.sum(ok) < self.min_cluster_size:
                        #~ self.log('!!!!!!! Not save trash!!!!!! np.sum(ok)', np.sum(ok))
                        #~ force_next_chan = None
                        #~ chan_visited.append(actual_chan)
                        #~ final_label = None
                        #~ candidate_labels_l0 = None
                        #~ self._plot_debug(actual_chan, ind_l0, features_l0, labels_l0, possible_labels_l0, candidate_labels_l0, final_label)
                        #~ continue
                    #~ labels_l0[ok] = 0
                    #~ labels_l0[~ok] = -1
                    #~ possible_labels_l0 = np.array([0], dtype='int64')
                #~ else:
                    #~ self.log('len(possible_labels_l0) == 0 AND diptest pval', pval, 'no cluster at all')
                    #~ if self.dense_mode:
                        #~ self.log('no more with dense mode')
                        #~ break
                    #~ else:
                        #~ force_next_chan = None
                        #~ chan_visited.append(actual_chan)
                        #~ #TODO dip test ????
                        #~ self.log('len(possible_labels_l0) == 0')
                        #~ final_label = None
                        #~ self._plot_debug(actual_chan, ind_l0, features_l0, labels_l0, possible_labels_l0, candidate_labels_l0, final_label)
                        #~ continue



            #~ self.debug_plot = iloop>20
            
            self.log('possible_labels_l0', possible_labels_l0)
            
            self.log('labels_l0 == -1 size', labels_l0[labels_l0==-1].size)
            

            # explore candidate
            possible_labels_l0, candidate_mask_l0, elsewhere_mask_l0, best_chan_peak_values, best_chan,\
                    peak_is_aligned, peak_is_on_chan, local_centroids = self.check_candidate_labels(ind_l0, labels_l0, actual_chan)
            
            
            self.log('possible_labels_l0', possible_labels_l0)
            candidate_labels_l0 = possible_labels_l0[candidate_mask_l0]
            self.log('candidate_labels_l0', candidate_labels_l0)
            candidate_labels_elsewhere_l0 = possible_labels_l0[elsewhere_mask_l0]
            self.log('candidate_labels_elsewhere_l0', candidate_labels_elsewhere_l0)
            
            
            if len(possible_labels_l0) == 0:
                self.log('len(possible_labels_l0) == 0 impossible')
                raise()
            
            if np.sum(elsewhere_mask_l0):
                ind_best_elsewhere = np.argmax(best_chan_peak_values[elsewhere_mask_l0])
                elsewhere_peak_val = best_chan_peak_values[elsewhere_mask_l0][ind_best_elsewhere]
            else:
                elsewhere_peak_val = None
            
            if np.sum(candidate_mask_l0):
                ind_best = np.argmax(best_chan_peak_values[candidate_mask_l0])
                peak_val = best_chan_peak_values[candidate_mask_l0][ind_best]
            else:
                peak_val = None
            
            self.log('peak_val', peak_val, 'elsewhere_peak_val', elsewhere_peak_val)
            
            final_label_elsewhere = None
            if elsewhere_peak_val is not None:
                # TODO put in params
                ratio_peak_force_explore = 1.5
                if (peak_val is not None and elsewhere_peak_val > peak_val*ratio_peak_force_explore) or peak_val is None:
                    # there is a better option elsewhere
                    force_next_chan = best_chan[elsewhere_mask_l0][ind_best_elsewhere]
                    if len(chan_visited)>0 and force_next_chan == chan_visited[-1]:
                        # last vist is the next one so lets avoid ping pong
                        if peak_val is None:
                            # keep actual chan anyway
                            final_label_elsewhere =possible_labels_l0[elsewhere_mask_l0][ind_best_elsewhere]
                            self.log('Best label other channel', force_next_chan, 'label', final_label_elsewhere)
                        else:
                            force_next_chan = None
                    
                    elif force_next_chan in chan_visited:
                        # but not last one
                        if peak_val is None:
                            self.log('Impossible to force_next_chan!!!!!!!!!!!!!!!', force_next_chan, 'actual_chan', actual_chan)
                            
                            force_next_chan = None
                            chan_visited.append(actual_chan)
                            
                            
                            final_label = None
                            self._plot_debug(actual_chan, ind_l0, features_l0, labels_l0, possible_labels_l0, candidate_labels_l0, final_label)
                            continue
                        else:
                            force_next_chan = None
                            # test peak_val
                    else:
                        # elsewhere have better option and not visisted yet
                        chan_visited.append(actual_chan)
                        self.log('force_next_chan', force_next_chan, 'actual_chan', actual_chan)
                        
                        final_label = None
                        self._plot_debug(actual_chan, ind_l0, features_l0, labels_l0, possible_labels_l0, candidate_labels_l0, final_label)
                        continue
            
            if peak_val is None and final_label_elsewhere is None:
                # no candidate and no elsewhere
                force_next_chan = None
                chan_visited.append(actual_chan)
                final_label = None
                self._plot_debug(actual_chan, ind_l0, features_l0, labels_l0, possible_labels_l0, candidate_labels_l0, final_label)
                continue                
            
            if final_label_elsewhere is not None:
                final_label = final_label_elsewhere
            else:
                final_label = possible_labels_l0[candidate_mask_l0][ind_best]
            
            self.log('final_label', final_label, 'peak_val', peak_val)
            
            # remove trash : TODO
            #~ if not self.dense_mode:
            #~ if len(possible_labels_l0) > 2:
            if True:
                # put -1 labels to trash
                ind_trash_label = ind_l0[labels_l0 == -1]
                cluster_labels[ind_trash_label] = -1
                self.log('trash -1 n=', ind_trash_label.shape)
                
                #~ # put peak_is_on_chan=True / aligned False
                #~ bad_labels = possible_labels_l0[~peak_is_aligned & peak_is_on_chan]
                # put aligned False
                bad_labels = possible_labels_l0[~peak_is_aligned]
                ind_trash_label = ind_l0[np.in1d(labels_l0, bad_labels)]
                self.log('trash bad_labels for trash', bad_labels, 'n=', ind_trash_label.size)
                cluster_labels[ind_trash_label] = -1
                


            ind_new_label = ind_l0[final_label == labels_l0]
            
            self.log('ind_new_label.shape', ind_new_label.shape)

            cluster_labels[cluster_labels>=k] += 1
            cluster_labels[ind_new_label] -= 1
            
            k += 1
            #~ chan_visited = []
            force_next_chan = None
            
            self._plot_debug(actual_chan, ind_l0, features_l0, labels_l0, possible_labels_l0, candidate_labels_l0, final_label)
            
            continue



            self._plot_debug(actual_chan, ind_l0, channel_features_l0, labels_l0, possible_labels_l0, ind_l2, labels_l2, possible_labels_l2, final_label_l2)
            
        self.log('END loop', np.sum(cluster_labels==-1))
        
        return cluster_labels


    def _plot_debug(self, actual_chan, ind_l0, features_l0, labels_l0, possible_labels_l0, candidate_labels_l0, final_label):
        if not self.debug_plot:
            return
        
        import matplotlib.pyplot as plt
        from .matplotlibplot import plot_waveforms_density
        fig, axs = plt.subplots(ncols=3, nrows=2)
        
        colors = plt.cm.get_cmap('Set3', len(possible_labels_l0))
        colors = {possible_labels_l0[l]:colors(l) for l in range(len(possible_labels_l0))}
        colors[-1] = 'k'
        
        ax = axs[0, 0]
        
        ax.set_title('actual_chan '+str(actual_chan))
        
        #~ if self.dense_mode:
            #~ mask_feat = slice(None)
        #~ else:
            #~ mask_feat = self.mask_feat_per_adj[actual_chan]
        local_wf_features = features_l0
        #~ print('local_wf_features.shape', local_wf_features.shape)
        
        #~ local_wf_features = features_l0
        
        if possible_labels_l0 is not None:
            for l, label in enumerate([-1] + possible_labels_l0.tolist()):
                #~ if label<0: 
                    #~ continue
                sel = label == labels_l0
                if np.sum(sel) ==0:
                    continue
                m = local_wf_features[sel].mean(axis=0)

                ind, = np.nonzero(sel)

                #~ if ind.size> 200:
                    #~ ind = ind[:200]
                if ind.size> self.max_per_cluster_for_median:
                    ind = ind[:self.max_per_cluster_for_median]

                color=colors[label]
                
                ax.plot(local_wf_features[ind].T, color=color, alpha=0.3)
                
                if label>=0:
                    ax.plot(m, color=color, alpha=1)
        
        ax = axs[0, 1]
        #~ n_components = min(local_wf_features.shape[1], self.n_components_local_pca)
        #~ pca =  sklearn.decomposition.IncrementalPCA(n_components=n_components, whiten=True)
        n_components = min(local_wf_features.shape[1]-1, self.n_components_local_pca)
        pca =  sklearn.decomposition.TruncatedSVD(n_components=n_components)
        reduced_features_l0 = pca.fit_transform(local_wf_features)
        #~ print('reduced_features_l0.shape', reduced_features_l0.shape)
        if possible_labels_l0 is not None:
            for l, label in enumerate([-1] + possible_labels_l0.tolist()):
                sel = label == labels_l0
                if np.sum(sel) ==0:
                    continue
                color=colors[label]
                #~ print('label', label, 'color', color)
                ax.scatter(reduced_features_l0[:, 0][sel], reduced_features_l0[:, 1][sel], color=color, s=2)

        ax = axs[1, 1]
        if possible_labels_l0 is not None:
            for l, label in enumerate([-1] + possible_labels_l0.tolist()):
                sel = label == labels_l0
                if np.sum(sel) ==0:
                    continue
                ind, = np.nonzero(sel)
                #~ if ind.size> 200:
                    #~ ind = ind[:200]
                if ind.size> self.max_per_cluster_for_median:
                    ind = ind[:self.max_per_cluster_for_median]
                    
                color=colors[label]
                ax.plot(reduced_features_l0[ind].T, color=color, alpha=0.3, lw=1)


                
        
        
        ax = axs[1, 0]
        adjacency = self.channel_adjacency[actual_chan]
        #~ print('plot actual_chan', actual_chan, 'adjacency', adjacency, )
        #~ wf_adj = self.waveforms[:,:, adjacency][ind_l0, :, :]
        wf_adj = self.cc.get_some_waveforms(self.peak_index[ind_l0], channel_indexes=adjacency)
        
        
        if possible_labels_l0 is not None:
            for l, label in enumerate([-1] + possible_labels_l0.tolist()):
                sel = label == labels_l0
                if np.sum(sel) ==0:
                    continue
                m = wf_adj[sel].mean(axis=0)
                ind, = np.nonzero(sel)

                #~ if ind.size> 200:
                    #~ ind = ind[:200]
                
                if ind.size> self.max_per_cluster_for_median:
                    ind = ind[:self.max_per_cluster_for_median]
                
                color=colors[label]
                
                ax.plot(wf_adj[ind].swapaxes(1,2).reshape(ind.size, -1).T, color=color, alpha=0.3)
                
                if label>=0:
                    ax.plot(m.T.flatten(), color=color, alpha=1, lw=2)
            
            
            for c, chan in enumerate(adjacency):
            #~ ind = adjacency.tolist().index(actual_chan)
                if self.dense_mode:
                    lw = 1
                else:
                    if chan == actual_chan:
                        lw = 2
                    else:
                        lw = 1
                ax.axvline(self.width * c - self.n_left, color='m', lw=lw)

        ax = axs[0, 2]
        if final_label is not None:
            
            adjacency = self.channel_adjacency[actual_chan]
            sel = final_label == labels_l0
            
            ax.set_title('keep size {}'.format(np.sum(sel)))
            
            #~ wf_adj = self.waveforms[:,:, adjacency][ind_l0, :, :][sel]
            wf_adj = self.cc.get_some_waveforms(self.peak_index[ind_l0], channel_indexes=adjacency)[sel]
            
            m = wf_adj.mean(axis=0)

            if wf_adj.shape[0] > 400:
                wf_adj = wf_adj[:400, :, :]
            
            color=colors[final_label]
            ax.plot(wf_adj.swapaxes(1,2).reshape(wf_adj.shape[0], -1).T, color=color, alpha=0.1)
            ax.plot(m.T.flatten(), color=color, alpha=1, lw=2)

        ax = axs[1, 2]
        if final_label is not None:
            sel = -1 == labels_l0
            
            ax.set_title('trash size {}'.format(np.sum(sel)))
            
            if np.sum(sel) > 0:
                #~ wf_adj = self.waveforms[:,:, adjacency][ind_l0, :, :][sel]
                wf_adj = self.cc.get_some_waveforms(self.peak_index[ind_l0], channel_indexes=adjacency)[sel]
                
                if wf_adj.shape[0] > 400:
                    wf_adj = wf_adj[:400, :, :]
                ax.plot(wf_adj.swapaxes(1,2).reshape(wf_adj.shape[0], -1).T, color='k', alpha=0.1)

            for c, chan in enumerate(adjacency):
            #~ ind = adjacency.tolist().index(actual_chan)
                if self.dense_mode:
                    lw = 1
                else:
                    if chan == actual_chan:
                        lw = 2
                    else:
                        lw = 1
                ax.axvline(self.width * c - self.n_left, color='m', lw=lw)

        

        plt.show()
        
        #~ self.debug_plot = False
        
        return




import scipy.signal
import numpy as np

from .cltools import HAVE_PYOPENCL, OpenCL_Helper
if HAVE_PYOPENCL:
    import pyopencl
    mf = pyopencl.mem_flags


#~ from pyacq.dsp.overlapfiltfilt import SosFiltfilt_Scipy
from .tools import FifoBuffer, median_mad


def offline_signal_preprocessor(sigs, sample_rate, common_ref_removal=True,
        highpass_freq=300., lowpass_freq=None, output_dtype='float32', normalize=True, **unused):
    #cast
    sigs = sigs.astype(output_dtype)
    
    #filter
    if highpass_freq is not None:
        b, a = scipy.signal.iirfilter(5, highpass_freq/sample_rate*2, analog=False,
                                        btype = 'highpass', ftype = 'butter', output = 'ba')
        filtered_sigs = scipy.signal.filtfilt(b, a, sigs, axis=0)
    else:
        filtered_sigs = sigs.copy()
    
    if lowpass_freq is not None:
        b, a = scipy.signal.iirfilter(5, lowpass_freq/sample_rate*2, analog=False,
                                        btype = 'lowpass', ftype = 'butter', output = 'ba')
        filtered_sigs = scipy.signal.filtfilt(b, a, filtered_sigs, axis=0)
        

    # common reference removal
    if common_ref_removal:
        filtered_sigs = filtered_sigs - np.median(filtered_sigs, axis=1)[:, None]
    
    # normalize
    if normalize:
        #~ med = np.median(filtered_sigs, axis=0)
        #~ mad = np.median(np.abs(filtered_sigs-med),axis=0)*1.4826
        med, mad = median_mad(filtered_sigs, axis=0)
        
        normed_sigs = (filtered_sigs - med)/mad
    else:
        normed_sigs = filtered_sigs
    
    return normed_sigs.astype(output_dtype)


def estimate_medians_mads_after_preprocesing(sigs, sample_rate, **params):
    params2 = dict(params)
    params2['normalize'] = False
    
    filtered_sigs = offline_signal_preprocessor(sigs, sample_rate, **params2)
    med, mad = median_mad(filtered_sigs, axis=0)
    return med, mad
    
    
    


class SignalPreprocessor_base:
    def __init__(self,sample_rate, nb_channel, chunksize, input_dtype):
        self.sample_rate = sample_rate
        self.nb_channel = nb_channel
        self.chunksize = chunksize
        self.input_dtype = input_dtype

    
    def change_params(self, common_ref_removal=True,
                                            highpass_freq=300.,
                                            lowpass_freq=None,
                                            smooth_size=0,
                                            output_dtype='float32', 
                                            normalize=True,
                                            pad_width = None,
                                            signals_medians=None, signals_mads=None):
                
        self.signals_medians = signals_medians
        self.signals_mads = signals_mads
        
        self.common_ref_removal = common_ref_removal
        self.highpass_freq = highpass_freq
        self.lowpass_freq = lowpass_freq
        self.smooth_size = int(smooth_size)
        self.output_dtype = np.dtype(output_dtype)
        self.normalize = normalize
        self.pad_width = pad_width
        
        # set default pad_width if none is provided
        if self.pad_width is None or self.pad_width<=0:
            assert self.highpass_freq is not None, 'pad_width=None needs a highpass_freq'
            self.pad_width = int(self.sample_rate/self.highpass_freq*3)
            #~ print('self.pad_width', self.pad_width)
        
        self.chunksize_1pad = self.chunksize + self.pad_width
        self.chunksize_2pad = self.chunksize + 2 * self.pad_width
        #~ print('self.pad_width', self.pad_width)
        #~ print('self.chunksize_1pad', self.chunksize_1pad)
        #~ assert self.chunksize_1pad>self.chunksize
        
        self.coefficients = np.zeros((0, 6))
        
        nyquist = self.sample_rate/2.
        
        if self.highpass_freq is not None:
            if self.highpass_freq>0 and self.highpass_freq<nyquist:
                coeff_hp = scipy.signal.iirfilter(5, highpass_freq/self.sample_rate*2, analog=False,
                                        btype = 'highpass', ftype = 'butter', output = 'sos')
                self.coefficients = np.concatenate((self.coefficients, coeff_hp))
        
        if self.lowpass_freq is not None:
            if self.lowpass_freq>0 and self.lowpass_freq<nyquist:
            #~ if self.lowpass_freq>(self.sample_rate/2.):
                #~ self.lowpass_freq=(self.sample_rate/2.01)
                coeff_lp = scipy.signal.iirfilter(5, lowpass_freq/self.sample_rate*2, analog=False,
                                        btype = 'lowpass', ftype = 'butter', output = 'sos')
                self.coefficients = np.concatenate((self.coefficients, coeff_lp))
        
        if self.smooth_size>0:
            b0 = (1./3)**.5
            b1 = (1-b0)
            b2 = 0.
            coeff_smooth = np.array([[b0, b1, b2, 1,0,0]], dtype=self.output_dtype)
            coeff_smooth = np.tile(coeff_smooth, (self.smooth_size, 1))
            self.coefficients = np.concatenate((self.coefficients, coeff_smooth))
        
        
        
        
        if self.coefficients.shape[0]==0:
            #this is the null filter
            self.coefficients = np.array([[1, 0, 0, 1,0,0]], dtype=self.output_dtype)
        
        self.nb_section =self. coefficients.shape[0]
        self.forward_buffer = FifoBuffer((self.chunksize_1pad, self.nb_channel), self.output_dtype)
        self.zi = np.zeros((self.nb_section, 2, self.nb_channel), dtype= self.output_dtype)
        
        #~ print('self.normalize', self.normalize)
        if self.normalize:
            assert self.signals_medians is not None
            assert self.signals_mads is not None

    def process_buffer(self, data):
        # used for offline processing when parralisation  is possible
        raise(NotImplmentedError)

    def initialize_stream(self):
        # must be for each new segment when index 
        # start back
        raise(NotImplmentedError)
    
    def process_buffer_stream(self, pos, data):
        # used in real time mode when chunk are given one after another
        raise(NotImplmentedError)
    


class SignalPreprocessor_Numpy(SignalPreprocessor_base):
    """
    This apply chunk by chunk on a multi signal:
       * baseline removal
       * hight pass filtfilt
       * normalize (optional)
    
    """
    def process_buffer(self, data):
        data = data.astype(self.output_dtype)
        processed_data = scipy.signal.sosfiltfilt(self.coefficients, data, axis=0)
        # TODO find why sosfiltfilt reverse strides!!!
        processed_data = np.ascontiguousarray(processed_data, dtype=self.output_dtype)
        # removal ref
        if self.common_ref_removal:
            processed_data -= np.median(processed_data, axis=1)[:, None]

        #normalize
        if self.normalize:
            processed_data -= self.signals_medians
            processed_data /= self.signals_mads

        return processed_data
    
    def process_buffer_stream(self, pos, data):
        # TODO rewrite this with self.process_buffer()
        
        #Online filtfilt
        chunk = data.astype(self.output_dtype)
        
        forward_chunk_filtered, self.zi = scipy.signal.sosfilt(self.coefficients, chunk, zi=self.zi, axis=0)
        forward_chunk_filtered = forward_chunk_filtered.astype(self.output_dtype)
        
        self.forward_buffer.new_chunk(forward_chunk_filtered, index=pos)
        
        backward_chunk = self.forward_buffer.buffer
        backward_filtered = scipy.signal.sosfilt(self.coefficients, backward_chunk[::-1, :], zi=None, axis=0)
        backward_filtered = backward_filtered[::-1, :]
        backward_filtered = backward_filtered.astype(self.output_dtype)
        
        pos2 = pos-self.pad_width
        if pos2<0:
            return None, None
        
        i1 = self.chunksize_1pad-self.pad_width-chunk.shape[0]
        i2 = self.chunksize

        assert i1<i2
        data2 = backward_filtered[i1:i2]
        if (pos2-data2.shape[0])<0:
            data2 = data2[data2.shape[0]-pos2:]
        
        # removal ref
        if self.common_ref_removal:
            data2 -= np.median(data2, axis=1)[:, None]
        
        #normalize
        if self.normalize:
            data2 -= self.signals_medians
            data2 /= self.signals_mads
        
        return pos2, data2
    
    def initialize_stream(self):
        self.forward_buffer.reset()
        self.zi[:] = 0
    
        
        



class SignalPreprocessor_OpenCL(SignalPreprocessor_base, OpenCL_Helper):
    """
    Implementation in OpenCL depending on material and nb_channel
    this can lead to a smal speed improvement...
    
    """
    def __init__(self,sample_rate, nb_channel, chunksize, input_dtype):
        SignalPreprocessor_base.__init__(self,sample_rate, nb_channel, chunksize, input_dtype)
    
    def _check_data(self, data):
        if not data.flags['C_CONTIGUOUS'] or data.dtype!=self.output_dtype:
            data = np.ascontiguousarray(data, dtype=self.output_dtype)

        return data
        
    
    def process_buffer(self, data):
        data = self._check_data(data)
        #~ print(data.shape, self.chunksize, self.chunksize_2pad, self.pad_width)
        #~ assert data.shape[0] == self.chunksize_2pad
        if data.shape[0] == self.chunksize_2pad:
            # OK
            unpad = 0
        elif data.shape[0] < self.chunksize_2pad:
            # put some zero
            unpad = self.chunksize_2pad - data.shape[0]
            data_pad = np.zeros((self.chunksize_2pad, data.shape[1]), dtype=data.dtype)
            #~ print('Apply a data pad')
            data = data_pad
        else:
            raise ValueError(f'data have wring shape{data.shape[0]} { self.chunksize_2pad}')
        
        event = pyopencl.enqueue_copy(self.queue,  self.input_2pad_cl, data)

        event = self.kern_forward_backward_filter(self.queue,  (self.nb_channel,), (self.nb_channel,),
                            self.input_2pad_cl, self.coefficients_cl, self.zi1_cl, self.zi2_cl,
                            self.signals_medians_cl, self.signals_mads_cl,  self.output_2pad_cl)
        #~ event.wait()
        
        event = pyopencl.enqueue_copy(self.queue,  self.output_2pad, self.output_2pad_cl)
        event.wait()
        
        data2 = self.output_2pad.copy()
        
        if self.common_ref_removal:
            # at the moment common_ref_removal is done on CPU
            # and so to avoid transfer normalize is also done on CPU
            #TODO implement OpenCL for removal ref
            if self.common_ref_removal:
                data2 -= np.median(data2, axis=1)[:, None]
            
            #normalize
            if self.normalize:
                # OpenCL for this when no common_ref_removal
                data2 -= self.signals_medians
                data2 /= self.signals_mads
        
        if unpad > 0:
            data2 = data2[:-unpad, :]
        
        return data2
    
    def process_buffer_stream(self, pos, data):
        
        assert data.shape[0]==self.chunksize
        
        data = self._check_data(data)
        
        #Online filtfilt
        event = pyopencl.enqueue_copy(self.queue,  self.input_cl, data)
        event = self.kern_stream_forward_backward_filter(self.queue,  (self.nb_channel,), (self.nb_channel,),
                            self.input_cl, self.coefficients_cl, self.zi1_cl, self.zi2_cl,
                            self.fifo_input_backward_cl, self.signals_medians_cl, self.signals_mads_cl,  self.output_backward_cl)
        event.wait()
        
         
        
        
        #~ event.wait()
        
        start = pos-self.chunksize_1pad
        if start<-self.pad_width:
            return None, None
        
        pos2 = pos-self.pad_width
        

        event = pyopencl.enqueue_copy(self.queue,  self.output_backward, self.output_backward_cl)
        if start>0:
            data2 = self.output_backward[:self.chunksize, :]
        else:
            data2 = self.output_backward[self.pad_width:self.chunksize, :]
        data2 = data2.copy()
        
        if self.common_ref_removal:
            # at the moment common_ref_removal is done on CPU
            # and so to avoid transfer normalize is also done on CPU
            #TODO implement OpenCL for removal ref
            if self.common_ref_removal:
                data2 -= np.median(data2, axis=1)[:, None]
            
            #normalize
            if self.normalize:
                # OpenCL for this when no common_ref_removal
                data2 -= self.signals_medians
                data2 /= self.signals_mads
            
        return pos2, data2        
        
        
    def change_params(self, **kargs):

        cl_platform_index=kargs.pop('cl_platform_index', None)
        cl_device_index=kargs.pop('cl_device_index', None)
        ctx=kargs.pop('ctx', None)
        queue=kargs.pop('queue', None)
        OpenCL_Helper.initialize_opencl(self,cl_platform_index=cl_platform_index, cl_device_index=cl_device_index, ctx=ctx, queue=queue)
        
        SignalPreprocessor_base.change_params(self, **kargs)
        assert self.output_dtype=='float32', 'SignalPreprocessor_OpenCL support only float32 at the moment'
        assert self.pad_width<self.chunksize, 'OpenCL fifo work only for self.pad_width<self.chunksize'
        
        
        
        self.coefficients = np.ascontiguousarray(self.coefficients, dtype=self.output_dtype)
        #~ print(self.coefficients.shape)
        
        # this is for stream processing
        self.zi1 = np.zeros((self.nb_channel, self.nb_section, 2), dtype= self.output_dtype)
        self.zi2 = np.zeros((self.nb_channel, self.nb_section, 2), dtype= self.output_dtype)
        self.output_forward = np.zeros((self.chunksize, self.nb_channel), dtype= self.output_dtype)
        self.fifo_input_backward = np.zeros((self.chunksize_1pad, self.nb_channel), dtype= self.output_dtype)
        self.output_backward = np.zeros((self.chunksize_1pad, self.nb_channel), dtype= self.output_dtype)
        
        #GPU buffers
        self.coefficients_cl = pyopencl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.coefficients)
        self.zi1_cl = pyopencl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=self.zi1)
        self.zi2_cl = pyopencl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=self.zi2)
        self.input_cl = pyopencl.Buffer(self.ctx, mf.READ_WRITE, size=self.output_forward.nbytes)
        self.output_forward_cl = pyopencl.Buffer(self.ctx, mf.READ_WRITE, size=self.output_forward.nbytes)
        self.fifo_input_backward_cl = pyopencl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.fifo_input_backward)
        self.output_backward_cl = pyopencl.Buffer(self.ctx, mf.READ_WRITE, size=self.output_backward.nbytes)
        if self.signals_medians is not None:
            self.signals_medians_cl = pyopencl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.signals_medians)
            self.signals_mads_cl = pyopencl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.signals_mads)
        else:
            self.signals_medians_cl = pyopencl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=np.zeros(self.nb_channel, dtype= self.output_dtype))
            self.signals_mads_cl = pyopencl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=np.zeros(self.nb_channel, dtype= self.output_dtype))
        
        # this is for offline processing
        self.input_2pad = np.zeros((self.chunksize_2pad, self.nb_channel), dtype= self.output_dtype)
        self.output_2pad = np.zeros((self.chunksize_2pad, self.nb_channel), dtype= self.output_dtype)
        self.input_2pad_cl = pyopencl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=self.input_2pad)
        self.output_2pad_cl = pyopencl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=self.output_2pad)


        #CL prog
        if not self.common_ref_removal and  self.normalize:
            extra_code_nomalize = _extra_code_nomalize
            extra_code_nomalize2 = _extra_code_nomalize2
        else:
            extra_code_nomalize = ''
            extra_code_nomalize2 = ''
        
        kernel_formated = processor_kernel%dict(chunksize=self.chunksize, chunksize_1pad=self.chunksize_1pad,
                        chunksize_2pad=self.chunksize_2pad,
                        pad_width=self.pad_width, nb_section=self.nb_section, nb_channel=self.nb_channel, 
                        extra_code_nomalize=extra_code_nomalize, extra_code_nomalize2=extra_code_nomalize2)
        #~ print(kernel_formated)
        prg = pyopencl.Program(self.ctx, kernel_formated)
        self.opencl_prg = prg.build(options='-cl-mad-enable')
        
        self.max_wg_size = self.ctx.devices[0].get_info(pyopencl.device_info.MAX_WORK_GROUP_SIZE)


        self.kern_stream_forward_backward_filter = getattr(self.opencl_prg, 'stream_forward_backward_filter')
        self.kern_forward_backward_filter = getattr(self.opencl_prg, 'forward_backward_filter')

    def initialize_stream(self):
        self.output_forward[:] = 0
        event = pyopencl.enqueue_copy(self.queue,  self.output_backward_cl, self.output_backward)
        event.wait()
        
        self.zi1[:] = 0
        event = pyopencl.enqueue_copy(self.queue,  self.zi1_cl, self.zi1)
        event.wait()

        self.zi2[:] = 0
        event = pyopencl.enqueue_copy(self.queue,  self.zi2_cl, self.zi2)
        event.wait()


processor_kernel = """
#define chunksize %(chunksize)d
#define chunksize_1pad %(chunksize_1pad)d
#define chunksize_2pad %(chunksize_2pad)d
#define pad_width %(pad_width)d
#define nb_section %(nb_section)d
#define nb_channel %(nb_channel)d


__kernel void sos_filter(__global  float *input, __global  float *output, __constant  float *coefficients, 
                                                                        __global float *zi, int local_chunksize, int direction, int out_offset_index) {

    int chan = get_global_id(0); //channel indice
    
    int offset_filt2;  //offset channel within section
    int offset_zi = chan*nb_section*2;
    
    int idx;

    float w0, w1,w2;
    float res;
    
    for (int section=0; section<nb_section; section++){
    
        //offset_filt2 = chan*nb_section*6+section*6;
        offset_filt2 = section*6;
        
        w1 = zi[offset_zi+section*2+0];
        w2 = zi[offset_zi+section*2+1];
        
        for (int s=0; s<local_chunksize;s++){
            
            if (direction==1) {idx = s*nb_channel+chan;}
            else if (direction==-1) {idx = (local_chunksize-s-1)*nb_channel+chan;}
            
            if (section==0)  {w0 = input[idx];}
            else {w0 = output[idx+out_offset_index];}
            
            w0 -= coefficients[offset_filt2+4] * w1;
            w0 -= coefficients[offset_filt2+5] * w2;
            res = coefficients[offset_filt2+0] * w0 + coefficients[offset_filt2+1] * w1 +  coefficients[offset_filt2+2] * w2;
            w2 = w1; w1 =w0;
            
            output[idx+out_offset_index] = res;
        }
        
        zi[offset_zi+section*2+0] = w1;
        zi[offset_zi+section*2+1] = w2;

    }
   
}


__kernel void stream_forward_backward_filter(__global  float *input,
                                                            __constant  float * coefficients,
                                                            __global float * zi1,
                                                            __global float * zi2,
                                                            __global  float *fifo_input_backward,
                                                            __global  float *signals_medians,
                                                            __global  float *signals_mads,
                                                            __global  float *output_backward){

    
    int chan = get_global_id(0); //channel indice


    //roll
    for (int s=0; s<pad_width;s++){
        fifo_input_backward[(s)*nb_channel+chan] = fifo_input_backward[(s+chunksize)*nb_channel+chan];
    }

    int out_offset_index = pad_width*nb_channel;
    sos_filter(input, fifo_input_backward, coefficients, zi1, chunksize, 1, out_offset_index);
    
    //set zi2 to zeros
    for (int s=0; s<nb_section;s++){
        zi2[chan*nb_section*2+s] = 0;
        zi2[chan*nb_section*2+s+1] = 0;
    }
    
    //filter backward
    sos_filter(fifo_input_backward, output_backward, coefficients, zi2, chunksize_1pad, -1, 0);
    
    // nomalize optional
    %(extra_code_nomalize)s

}

__kernel void forward_backward_filter(__global  float *input,
                                                            __constant  float * coefficients,
                                                            __global float * zi1,
                                                            __global float * zi2,
                                                            __global  float *signals_medians,
                                                            __global  float *signals_mads,
                                                            __global  float *output){

    
    int chan = get_global_id(0); //channel indice


    sos_filter(input, input, coefficients, zi1, chunksize_2pad, 1, 0);
    
    //filter backward
    sos_filter(input, output, coefficients, zi2, chunksize_2pad, -1, 0);
    
    // nomalize optional
    %(extra_code_nomalize2)s

}



"""

_extra_code_nomalize = """
    float v;
    for (int s=0; s<chunksize;s++){
        v = output_backward[(s)*nb_channel+chan];
        output_backward[(s)*nb_channel+chan] = (v - signals_medians[chan]) / signals_mads[chan];
    }
"""

_extra_code_nomalize2 = """
    float v;
    for (int s=0; s<chunksize_2pad;s++){
        v = output[(s)*nb_channel+chan];
        output[(s)*nb_channel+chan] = (v - signals_medians[chan]) / signals_mads[chan];
    }
"""



signalpreprocessor_engines = { 'numpy' : SignalPreprocessor_Numpy,
                                                'opencl' : SignalPreprocessor_OpenCL}



"""
This view is from taken from sklearn examples.
See http://scikit-learn.org: plot-kmeans-silhouette-analysis-py



"""
from .myqt import QT
import pyqtgraph as pg

import numpy as np
import matplotlib.cm
import matplotlib.colors


from .base import WidgetBase
from .tools import ParamDialog

class MyViewBox(pg.ViewBox):
    doubleclicked = QT.pyqtSignal()
    def mouseDoubleClickEvent(self, ev):
        self.doubleclicked.emit()
        ev.accept()
    def raiseContextMenu(self, ev):
        #for some reasons enableMenu=False is not taken (bug ????)
        pass



class Silhouette(WidgetBase):
    """
    **Silhouette**  display the silhouette score.
    
    Implemented with sklearn. Must compute metrics first.
    
    See:
      * `Silhouette wikipedia <https://en.wikipedia.org/wiki/Silhouette_(clustering)>`_
      * `Silhouette sklearn <http://scikit-learn.org/stable/auto_examples/cluster/plot_kmeans_silhouette_analysis.html#sphx-glr-auto-examples-cluster-plot-kmeans-silhouette-analysis-py>`_
    """
    
    _params = [
        ]
    
    def __init__(self, controller=None, parent=None):
        WidgetBase.__init__(self, parent=parent, controller=controller)
        
        self.layout = QT.QVBoxLayout()
        self.setLayout(self.layout)
        
        h = QT.QHBoxLayout()
        self.layout.addLayout(h)
        h.addWidget(QT.QLabel('<b>Silhouette</b>') )

        but = QT.QPushButton('settings')
        but.clicked.connect(self.open_settings)
        h.addWidget(but)

        self.graphicsview = pg.GraphicsView()
        self.layout.addWidget(self.graphicsview)
        
        self.alpha = 60
        
        self.initialize_plot()
        
        self.refresh()
        
    def on_params_changed(self):
        self.compute_slihouette()
        self.refresh()

    def initialize_plot(self):
        self.viewBox = MyViewBox()
        self.viewBox.doubleclicked.connect(self.open_settings)
        self.viewBox.disableAutoRange()
        
        self.plot = pg.PlotItem(viewBox=self.viewBox)
        self.graphicsview.setCentralItem(self.plot)
        self.plot.hideButtons()
    
    def refresh(self):
        self.plot.clear()
        silhouette_values = self.controller.spike_silhouette
        if silhouette_values is None:
            return
        
        if silhouette_values.shape != self.controller.spike_label.shape:
            return
        
        silhouette_avg = np.mean(silhouette_values)
        silhouette_by_labels = {}
        labels = self.controller.spike_label
        labels_list = np.unique(labels)
        for k in labels_list:
            v = silhouette_values[k==labels]
            v.sort()
            silhouette_by_labels[k] = v
        
        
        self.vline = pg.InfiniteLine(pos=silhouette_avg, angle = 90, movable = False, pen = '#FF0000')
        self.plot.addItem(self.vline)
        
        y_lower = 10
        cluster_visible = self.controller.cluster_visible
        visibles = [c for c, v in self.controller.cluster_visible.items() if v and c>=0]
        
        for k in visibles:
            if k not in silhouette_by_labels:
                continue
            v = silhouette_by_labels[k]
            
            color = self.controller.qcolors[k]
            color2 = QT.QColor(color)
            color2.setAlpha(self.alpha)
            
            y_upper = y_lower + v.size
            y_vect = np.arange(y_lower, y_upper)
            curve1 = pg.PlotCurveItem(np.zeros(v.size), y_vect, pen=color)
            curve2 = pg.PlotCurveItem(v, y_vect, pen=color)
            self.plot.addItem(curve1)
            self.plot.addItem(curve2)
            fill = pg.FillBetweenItem(curve1=curve1, curve2=curve2, brush=color2)
            self.plot.addItem(fill)
            
            txt = pg.TextItem( text='{}'.format(k), color='#FFFFFF', anchor=(0, 0.5), border=None)#, fill=pg.mkColor((128,128,128, 180)))
            self.plot.addItem(txt)
            txt.setPos(0, (y_upper+y_lower)/2.)
            
            y_lower = y_upper + 10

        
        self.plot.setXRange(-.5, 1.)
        self.plot.setYRange(0,y_lower)


    def on_spike_selection_changed(self):
        pass

    def on_spike_label_changed(self):
        #~ self.compute_slihouette()
        self.refresh()
    
    def on_colors_changed(self):
        self.refresh()
    
    def on_cluster_visibility_changed(self):
        self.refresh()
        



from .myqt import QT
import pyqtgraph as pg

import numpy as np
#import seaborn as sns


from .base import ControllerBase

from .. import labelcodes
from ..tools import make_color_dict

from ..peeler import _dtype_spike

spike_visible_modes = ['selected', 'all',  'collision']

#~ _dtype_spike = [('index', 'int64'), ('cluster_label', 'int64'), ('jitter', 'float64'),]

_dtype_complement = [('cell_label', 'int64'), ('channel', 'int64'), ('segment', 'int64'), ('visible', 'bool'),
                ('selected', 'bool')]


class PeelerController(ControllerBase):
    def __init__(self, parent=None, dataio=None, catalogue=None):
        ControllerBase.__init__(self, parent=parent)
        self.dataio = dataio
        self.catalogue = catalogue
        
        self.chan_grp = catalogue['chan_grp']
        self.nb_channel = self.dataio.nb_channel(self.chan_grp)
        self.channels = np.arange(self.nb_channel, dtype='int64')
        
        self.init_plot_attributes()
        self.update_visible_spikes()
    
    def init_plot_attributes(self):
        #concatenate all spikes for all segments
        all_spikes = []
        for i in range(self.dataio.nb_segment):
            local_spikes = self.dataio.get_spikes(seg_num=i, chan_grp=self.chan_grp)
            spikes = np.zeros(local_spikes.shape, dtype=_dtype_spike+_dtype_complement)
            for k, _ in _dtype_spike:
                spikes[k] = local_spikes[k]
            spikes['segment'] = i
            spikes['visible'] = spikes['cluster_label']>=0
            spikes['selected'] = False
            
            
            clusters = self.catalogue['clusters']
            cluster_labels = clusters['cluster_label']
            cell_labels = clusters['cell_label']
            
            #set cell_label <0 are the same >0 are converted for 'clusters' table in catalogue
            mask = spikes['cluster_label']<0
            spikes['cell_label'][mask] = spikes['cluster_label'][mask]
            
            mask = spikes['cluster_label']>=0
            # searchsorter wotk with ordered
            clus = clusters[clusters['cluster_label']>=0]
            order = np.argsort(clus['cluster_label'])
            clus = clus[order].copy()
            spike_cluster_index = np.searchsorted(clus['cluster_label'], spikes['cluster_label'][mask].copy())
            spikes['cell_label'][mask] = clus['cell_label'][spike_cluster_index]
            
            all_spikes.append(spikes)
        self._spikes = np.concatenate(all_spikes)
        
        self.nb_spike = int(self.spikes.size)
        
        self.cluster_labels = self.catalogue['clusters']['cluster_label']
        
        # set channel for each cluster
        for i, k in enumerate(self.cluster_labels):
            #~ chan = self.catalogue['extremum_channel'][i]
            chan = self.catalogue['clusters']['extremum_channel'][i]
            sel = self.spikes['cluster_label'] == k
            self.spikes['channel'][sel] = chan
        
        # compute sparse mask
        # TODO : put this mask in catalogue directly
        self.sparse_threshold = 1.5
        self.centroids_sparse_mask = np.zeros((self.cluster_labels.size, self.nb_channel), dtype='bool')
        for k in self.cluster_labels:
            ind = self.catalogue['label_to_index'][k] 
            median = self.catalogue['centers0'][ind, :, :]
            self.centroids_sparse_mask[ind, :] = np.any(np.abs(median) > self.sparse_threshold, axis=0)
        
        #~ self.cluster_labels = np.unique(self.spikes['cluster_label'])#TODO take from catalogue
        
        
        self.cluster_count = { k:np.sum(self.spikes['cluster_label']==k) for k in self.cluster_labels}
        
        
        self.cluster_visible = {k:k>=0 for k  in self.cluster_labels}

        
        #qt colors
        self.colors = make_color_dict(self.catalogue['clusters'])
        self.qcolors = {}
        for k, color in self.colors.items():
            r, g, b = color
            self.qcolors[k] = QT.QColor(r*255, g*255, b*255)
        
        self.spike_visible_mode = spike_visible_modes[0]
    
    def check_plot_attributes(self):
        for k in self.cluster_labels:
            if k not in self.cluster_visible:
                self.cluster_visible[k] = k>=0
        
        for k in list(self.cluster_visible.keys()):
            if k not in self.cluster_labels:
                self.cluster_visible.pop(k)
        
        #~ self.refresh_colors(reset=False)
    
    @property
    def have_sparse_template(self):
        return True

    @property
    def spikes(self):
        return self._spikes

    @property
    def spike_selection(self):
        return self.spikes['selected']

    @property
    def spike_segment(self):
        return self.spikes['segment']

    @property
    def spike_index(self):
        return self.spikes['index']

    @property
    def spike_channel(self):
        return self.spikes['index']

    @property
    def positive_cluster_labels(self):
        cluster_labels = self.clusters['cluster_label']
        return cluster_labels[cluster_labels>=0]

    @property
    def clusters(self):
        return self.catalogue['clusters']
    
    def get_waveforms_shape(self):
        shape = self.catalogue['centers0'].shape[1:]
        return shape
    
    def get_sparse_channels(self, label):
        ind = self.catalogue['label_to_index'][label] 
        chans,  = np.nonzero(self.centroids_sparse_mask[ind, :])
        return chans
    
    def get_common_sparse_channels(self, labels):
        inds = [self.catalogue['label_to_index'][label]  for label in labels]
        chans,  = np.nonzero(np.any(self.centroids_sparse_mask[inds, :], axis=0))
        return chans
    
    def get_waveform_centroid(self, label, metric, sparse=False, channels=None):
        if metric in ('mean', 'std', 'mad'):
            return None, None
        

        if label in self.catalogue['label_to_index']:
            i = self.catalogue['label_to_index'][label]
            wf = self.catalogue['centers0'][i, :, :].copy()

            if sparse:
                assert channels is None
                chans = self.get_sparse_channels(label)
                wf = wf[:, chans].copy()
            elif channels is not None:
                chans = channels
                wf = wf[:, chans].copy()
            else:
                chans = self.channels
                wf = wf.copy()
            
            return wf, chans
        else:
            return None, None

    def get_min_max_centroids(self):
        if self.catalogue['centers0'].shape[0]>0:
            wf_min = np.min(self.catalogue['centers0'])
            wf_max = np.max(self.catalogue['centers0'])
        else:
            wf_min = 0.
            wf_max = 0.
        return wf_min, wf_max

    def get_waveform_left_right(self):
        return self.catalogue['n_left'], self.catalogue['n_right']
    
    def get_threshold(self):
        threshold = self.catalogue['peak_detector_params']['relative_threshold']
        if self.catalogue['peak_detector_params']['peak_sign']=='-':
            threshold = -threshold
        return threshold

    def get_extremum_channel(self, label):
        if label in self.catalogue['label_to_index']:
            cluster_idx = self.catalogue['label_to_index'][label]
            #~ c = self.catalogue['extremum_channel'][cluster_idx]
            c = self.catalogue['clusters']['extremum_channel'][cluster_idx]
            return c
    
    def get_some_waveforms(self, seg_nums, peak_sample_indexes, channel_indexes):
        n_left, n_right = self.get_waveform_left_right()
        waveforms = self.dataio.get_some_waveforms(seg_nums=seg_nums, chan_grp=self.chan_grp, 
                            peak_sample_indexes=peak_sample_indexes,
                            n_left=n_left, n_right=n_right, channel_indexes=channel_indexes)
        return waveforms

    def change_spike_visible_mode(self, mode):
        assert mode in spike_visible_modes
        #~ print(mode)
        self.spike_visible_mode = mode
        
    def update_visible_spikes(self):
        #~ print('update_visible_spikes', self.spike_visible_mode)
        #~ ['selected', 'all',  'collision']
        if self.spike_visible_mode=='selected':
            visibles = np.array([k for k, v in self.cluster_visible.items() if v ])
            self.spikes['visible'][:] = np.in1d(self.spikes['cluster_label'], visibles)
        elif self.spike_visible_mode=='all':
            self.spikes['visible'][:] = True
        elif self.spike_visible_mode=='collision':
            self.spikes['visible'][:] = False
            d = np.diff(self.spikes['index'])
            labels0 = self.spikes['cluster_label'][:-1]
            labels1 = self.spikes['cluster_label'][1:]
            mask = (d>0) & (d< int(self.dataio.sample_rate * 0.001) ) & (labels0>0) & (labels1>0)
            ind, = np.nonzero(mask)
            self.spikes['visible'][ind] = True
            self.spikes['visible'][ind+1] = True
            
            
            
    
    def on_cluster_visibility_changed(self):
        #~ print('on_cluster_visibility_changed')
        self.update_visible_spikes()
        ControllerBase.on_cluster_visibility_changed(self)



from .myqt import QT
import pyqtgraph as pg

import numpy as np
import  itertools

from .base import WidgetBase
from .tools import ParamDialog



class PairList(WidgetBase):
    """
    **Pair list**  is an intuituitive list of pair of cluster : when you click on
    a pair this make visible only this 2 cluster on all others views.
    
    This help to validate in a fast way clusters.
    
    For convinience this paris can be filtered:
      * **all pairs** : all combinaison
      * **high similarity**: select only pairs that have similarity over a threshold (in settings)
      * **similarity amplitude ratio**! select only pairs that have "similarity ratio" over a threshold (in settings)
    
    And they can be sorted by:
      * label
      * similarity
      * ratio_similarity
    
    WIth a right click you can:
      * **merge** a pair
      * **tag same cell**, this keep 2 cluster but there tag as same cell.
    
    """
    _params = [{'name': 'threshold_similarity', 'type': 'float', 'value' :.9, 'step' : 0.01},
                    {'name': 'threshold_ratio_similarity', 'type': 'float', 'value' :.8, 'step' : 0.01},
                ]

    
    def __init__(self, controller=None, parent=None):
        WidgetBase.__init__(self, parent=parent, controller=controller)
        
        self.layout = QT.QVBoxLayout()
        self.setLayout(self.layout)

        #~ h = QT.QHBoxLayout()
        #~ self.layout.addLayout(h)
        self.combo_select = QT.QComboBox()
        #~ h.addWidget(QT.QLabel('Select'))
        #~ h.addWidget(self.combo_select)
        self.combo_select.addItems(['all pairs', 'similar amplitude ratio', 'high similarity']) #
        #~ self.combo_select.currentTextChanged.connect(self.refresh)
        #~ h.addStretch()

        h = QT.QHBoxLayout()
        self.layout.addLayout(h)
        h.addWidget(QT.QLabel('Sort by'))
        self.combo_sort = QT.QComboBox()
        self.combo_sort.addItems(['label', 'similarity', 'ratio_similarity'])
        self.combo_sort.currentIndexChanged.connect(self.refresh)
        h.addWidget(self.combo_sort)
        h.addStretch()
        
        
        #~ but = QT.QPushButton('settings')
        #~ self.layout.addWidget(but)
        #~ but.clicked.connect(self.open_settings)
        
        self.table = QT.QTableWidget(selectionMode=QT.QAbstractItemView.SingleSelection,
                                                        selectionBehavior=QT.QAbstractItemView.SelectRows)
        self.table.setContextMenuPolicy(QT.Qt.CustomContextMenu)
        self.layout.addWidget(self.table)
        self.table.itemSelectionChanged.connect(self.on_item_selection_changed)
        self.table.customContextMenuRequested.connect(self.open_context_menu)
        
        self.menu = QT.QMenu()
        act = self.menu.addAction('Merge')
        act.triggered.connect(self.do_merge)

        act = self.menu.addAction('Tag same cell')
        act.triggered.connect(self.do_tag_same_cell)
        
        self.refresh()
    
    def on_item_selection_changed(self):
        inds = self.table.selectedIndexes()
        if len(inds)!=self.table.columnCount():
            return
        k1, k2 = self.pairs[inds[0].row()]
        for k in self.controller.cluster_visible:
            self.controller.cluster_visible[k] = k in (k1, k2)
        
        self.cluster_visibility_changed.emit()

    def open_context_menu(self):
        self.menu.popup(self.cursor().pos())
    
    def do_merge(self):
        if len(self.table.selectedIndexes())==0:
            return
        ind = self.table.selectedIndexes()[0].row()
        
        label_to_merge = list(self.pairs[ind])
        self.controller.merge_cluster(label_to_merge)
        self.refresh()
        self.spike_label_changed.emit()
    
    def do_tag_same_cell(self):
        if len(self.table.selectedIndexes())==0:
            return
        ind = self.table.selectedIndexes()[0].row()
        
        label_to_merge = list(self.pairs[ind])
        self.controller.tag_same_cell(label_to_merge)
        self.refresh()
        self.cluster_tag_changed.emit()
        
    
    def refresh(self):
        self.table.clear()
        labels = ['cluster_label_1', 'cluster_label_2', 'cell_label_1', 'cell_label_2' , 'similarity', 'ratio_similarity']
        self.table.setColumnCount(len(labels))
        self.table.setHorizontalHeaderLabels(labels)
        self.table.setColumnWidth(0, 100)
        self.table.setColumnWidth(1, 100)
        
        #select
        mode = self.combo_select.currentText()
        if mode == 'all pairs':
            labels = self.controller.positive_cluster_labels
            #~ labels = labels[labels>=0]
            self.pairs = list(itertools.combinations(labels, 2))
        elif mode == 'similar amplitude ratio':
            self.pairs = self.controller.detect_similar_waveform_ratio(threshold=self.params['threshold_ratio_similarity'])
        elif mode == 'high similarity':
            self.pairs = self.controller.detect_high_similarity(threshold=self.params['threshold_similarity'])
        
        #sort
        mode = self.combo_sort.currentText()
        order = np.arange(len(self.pairs))
        if mode == 'label':
            pass
        elif mode == 'similarity':
            if self.controller.cluster_similarity is not None:
                order = []
                for r in range(len(self.pairs)):
                    k1, k2 = self.pairs[r]
                    ind1 = self.controller.positive_cluster_labels.tolist().index(k1)
                    ind2 = self.controller.positive_cluster_labels.tolist().index(k2)
                    order.append(self.controller.cluster_similarity[ind1, ind2])
                order = np.argsort(order)[::-1]
        elif mode == 'ratio_similarity':
            if self.controller.cluster_ratio_similarity is not None:
                order = []
                for r in range(len(self.pairs)):
                    k1, k2 = self.pairs[r]
                    ind1 = self.controller.positive_cluster_labels.tolist().index(k1)
                    ind2 = self.controller.positive_cluster_labels.tolist().index(k2)
                    order.append(self.controller.cluster_ratio_similarity[ind1, ind2])
                order = np.argsort(order)[::-1]
        self.pairs = [self.pairs[i] for i in order ]
        
        self.table.setRowCount(len(self.pairs))
        
        for r in range(len(self.pairs)):
            k1, k2 = self.pairs[r]
            ind1 = self.controller.positive_cluster_labels.tolist().index(k1)
            ind2 = self.controller.positive_cluster_labels.tolist().index(k2)
            
            for c, k in enumerate((k1, k2)):
                color = self.controller.qcolors.get(k, QT.QColor( 'white'))
                pix = QT.QPixmap(16,16)
                pix.fill(color)
                icon = QT.QIcon(pix)
                
                name = '{} (nb={})'.format(k, self.controller.cluster_count[k])
                item = QT.QTableWidgetItem(name)
                item.setFlags(QT.Qt.ItemIsEnabled|QT.Qt.ItemIsSelectable)
                self.table.setItem(r,c, item)
                item.setIcon(icon)
                
                
                cell_label = self.controller.cell_labels[self.controller.cluster_labels==k][0]
                name = '{}'.format(cell_label)
                item = QT.QTableWidgetItem(name)
                item.setFlags(QT.Qt.ItemIsEnabled|QT.Qt.ItemIsSelectable)
                self.table.setItem(r,c+2, item)
        
            if self.controller.cluster_similarity is not None:
                if self.controller.cluster_similarity.shape[0] == self.controller.positive_cluster_labels.size:
                    name = '{}'.format(self.controller.cluster_similarity[ind1, ind2])
                    item = QT.QTableWidgetItem(name)
                    item.setFlags(QT.Qt.ItemIsEnabled|QT.Qt.ItemIsSelectable)
                    self.table.setItem(r,4, item)

            if self.controller.cluster_ratio_similarity is not None:
                if self.controller.cluster_ratio_similarity.shape[0] == self.controller.positive_cluster_labels.size:
                    name = '{}'.format(self.controller.cluster_ratio_similarity[ind1, ind2])
                    item = QT.QTableWidgetItem(name)
                    item.setFlags(QT.Qt.ItemIsEnabled|QT.Qt.ItemIsSelectable)
                    self.table.setItem(r,5, item)
        
                


        

    def on_spike_selection_changed(self):
        pass

    def on_spike_label_changed(self):
        self.refresh()
    
    def on_colors_changed(self):
        self.refresh()
    
    def on_cluster_visibility_changed(self):
        pass




import numpy as np

import sklearn
import sklearn.decomposition
import sklearn.cluster
import sklearn.manifold
import sklearn.discriminant_analysis

from . import tools

import joblib

import time

def project_waveforms(method='pca_by_channel', catalogueconstructor=None, selection=None, **params):
    """
    If slection is None then the fit of projector is done on all some_peaks_index
    
    Otherwise the fit is done on the a subset of waveform force by slection bool mask
    
    selection is mask bool size all_peaks
    """
    cc = catalogueconstructor
    
    
    if method=='global_pca':
        projector = GlobalPCA(catalogueconstructor=cc, selection=selection, **params)
    elif method=='peak_max':
        projector = PeakMaxOnChannel(catalogueconstructor=cc, selection=selection, **params)
    elif method=='pca_by_channel':
        projector = PcaByChannel(catalogueconstructor=cc, selection=selection, **params)
    #~ elif method=='neighborhood_pca':
        #~ projector = NeighborhoodPca(waveforms, catalogueconstructor=catalogueconstructor, **params)
    elif method=='global_lda':
        projector = GlobalLDA(catalogueconstructor=cc, selection=selection, **params)
    else:
        raise NotImplementedError
    
    #~ features = projector.transform(waveforms2)
    features = projector.get_features(catalogueconstructor)
    channel_to_features = projector.channel_to_features
    return features, channel_to_features, projector


class GlobalPCA:
    def __init__(self, catalogueconstructor=None, selection=None, n_components=5, **params):

        cc = catalogueconstructor
        
        self.n_components = n_components
        
        self.waveforms = cc.get_some_waveforms()
        
        if selection is None:
            waveforms = self.waveforms
            #~ print('all selection', waveforms.shape[0])
        else:
            peaks_index, = np.nonzero(selection)
            waveforms = cc.get_some_waveforms(peaks_index=peaks_index)
            #~ print('subset selection', waveforms.shape[0])
        
        #~ print(waveforms.shape)
        flatten_waveforms = waveforms.reshape(waveforms.shape[0], -1)
        #~ self.pca =  sklearn.decomposition.IncrementalPCA(n_components=n_components, **params)
        self.pca =  sklearn.decomposition.TruncatedSVD(n_components=n_components, **params)
        self.pca.fit(flatten_waveforms)
        
        
        #In GlobalPCA all feature represent all channels
        self.channel_to_features = np.ones((cc.nb_channel, self.n_components), dtype='bool')

    def get_features(self, catalogueconstructor):
        features = self.transform(self.waveforms)
        del self.waveforms
        return features

    def transform(self, waveforms):
        flatten_waveforms = waveforms.reshape(waveforms.shape[0], -1)
        return self.pca.transform(flatten_waveforms)

class PeakMaxOnChannel:
    def __init__(self,  catalogueconstructor=None, selection=None, **params):
        if selection is not None:
            print('selection with PeakMaxOnChannel is a non sens')
            
        cc = catalogueconstructor
        
        #~ self.waveforms = waveforms
        # TODO something faster with only the max!!!!!
        self.waveforms = cc.get_some_waveforms()
        
        self.ind_peak = -catalogueconstructor.info['extract_waveforms']['n_left']
        #~ print('PeakMaxOnChannel self.ind_peak', self.ind_peak)
        
        
        #In full PeakMaxOnChannel one feature is one channel
        self.channel_to_features = np.eye(cc.nb_channel, dtype='bool')
    
    def get_features(self, catalogueconstructor):
        features = self.transform(self.waveforms)
        del self.waveforms
        return features
    
        
    def transform(self, waveforms):
        #~ print('ici', waveforms.shape, self.ind_peak)
        features = waveforms[:, self.ind_peak, : ].copy()
        return features



#~ Parallel(n_jobs=n_jobs)(delayed(count_match_spikes)(sorting1.get_unit_spike_train(u1),
                                                                                  #~ s2_spiketrains, delta_frames) for
                                                      #~ i1, u1 in enumerate(unit1_ids))

#~ def get_pca_one_channel(wf_chan, chan, thresh, n_left, n_components_by_channel, params):
    #~ print(chan)
    #~ pca = sklearn.decomposition.IncrementalPCA(n_components=n_components_by_channel, **params)
    #~ wf_chan = waveforms[:,:,chan]
    #~ print(wf_chan.shape)
    #~ print(wf_chan[:, -n_left].shape)
    #~ keep = np.any((wf_chan>thresh) | (wf_chan<-thresh))
    #~ keep = (wf_chan[:, -n_left]>thresh) | (wf_chan[:, -n_left]<-thresh)

    #~ if keep.sum() >=n_components_by_channel:
        #~ pca.fit(wf_chan[keep, :])
        #~ return pca
    #~ else:
        #~ return None


class PcaByChannel:
    def __init__(self, catalogueconstructor=None, selection=None, n_components_by_channel=3, adjacency_radius_um=200, **params):
        
        cc = catalogueconstructor
        
        thresh = cc.info['peak_detector']['relative_threshold']
        n_left = cc.info['extract_waveforms']['n_left']
        self.dtype = cc.info['internal_dtype']
        
        
        #~ self.waveforms = waveforms
        self.n_components_by_channel = n_components_by_channel
        self.adjacency_radius_um = adjacency_radius_um
        
        
        #~ t1 = time.perf_counter()
        if selection is None:
            peaks_index = cc.some_peaks_index
        else:
            peaks_index,  = np.nonzero(selection)
        
        some_peaks = cc.all_peaks[peaks_index]
        
        self.pcas = []
        for chan in range(cc.nb_channel):
        #~ for chan in range(20):
            #~ print('fit', chan)
            sel = some_peaks['channel'] == chan
            wf_chan = cc.get_some_waveforms(peaks_index=peaks_index[sel], channel_indexes=[chan])
            wf_chan = wf_chan[:, :, 0]
            #~ print(wf_chan.shape)
            
            if wf_chan.shape[0] - 1 > n_components_by_channel:
                #~ pca = sklearn.decomposition.IncrementalPCA(n_components=n_components_by_channel, **params)
                #~ print('PcaByChannel SVD')
                pca = sklearn.decomposition.TruncatedSVD(n_components=n_components_by_channel, **params)
                #~ print(wf_chan.shape)
                #~ print(np.sum(np.isnan(wf_chan)))
                #~ import matplotlib.pyplot as plt
                #~ fig, ax = plt.subplots()
                #~ ax.plot(wf_chan.T)
                #~ plt.show()
                try:
                    pca.fit(wf_chan)
                except ValueError:
                    pca = None
                    print('Error in PcaByChannel for channel {} maybe too noisy (Nan, Inf,)'.format(chan))
            else:
                pca = None
            self.pcas.append(pca)

        #~ t2 = time.perf_counter()
        #~ print('pca fit', t2-t1)
            


            
            #~ pca = get_pca_one_channel(waveforms, chan, thresh, n_left, n_components_by_channel, params)
            
        #~ n_jobs = -1
        #~ self.pcas = joblib.Parallel(n_jobs=n_jobs)(joblib.delayed(get_pca_one_channel)(waveforms, chan, thresh, n_components_by_channel, params) for chan in range(cc.nb_channel))
        

        #In full PcaByChannel n_components_by_channel feature correspond to one channel
        self.channel_to_features = np.zeros((cc.nb_channel, cc.nb_channel*n_components_by_channel), dtype='bool')
        for c in range(cc.nb_channel):
            self.channel_to_features[c, c*n_components_by_channel:(c+1)*n_components_by_channel] = True

    def get_features(self, catalogueconstructor):
        cc = catalogueconstructor
        
        nb = cc.some_peaks_index.size
        n = self.n_components_by_channel
        
        features = np.zeros((nb, cc.nb_channel*self.n_components_by_channel), dtype=self.dtype)
        
        some_peaks = cc.all_peaks[cc.some_peaks_index]

        if cc.mode == 'sparse':
            assert cc.info['peak_detector']['method'] == 'geometrical'
            #~ adjacency_radius_um = cc.info['peak_detector']['adjacency_radius_um']
            channel_adjacency = cc.dataio.get_channel_adjacency(chan_grp=cc.chan_grp, adjacency_radius_um=self.adjacency_radius_um)
        
        #~ t1 = time.perf_counter()
        for chan, pca in enumerate(self.pcas):
            if pca is None:
                continue
            #~ print('transform', chan)
            #~ sel = some_peaks['channel'] == chan
            
            if cc.mode == 'dense':
                wf_chan = cc.get_some_waveforms(peaks_index=cc.some_peaks_index, channel_indexes=[chan])
                wf_chan = wf_chan[:, :, 0]
                #~ print('dense', wf_chan.shape)
                features[:, chan*n:(chan+1)*n] = pca.transform(wf_chan)
            elif cc.mode == 'sparse':
                sel = np.in1d(some_peaks['channel'], channel_adjacency[chan])
                #~ print(chan, np.sum(sel))
                wf_chan = cc.get_some_waveforms(peaks_index=cc.some_peaks_index[sel], channel_indexes=[chan])
                wf_chan = wf_chan[:, :, 0]
                #~ print('sparse', wf_chan.shape)
                features[:, chan*n:(chan+1)*n][sel, :] = pca.transform(wf_chan)

        #~ t2 = time.perf_counter()
        #~ print('pca transform', t2-t1)
            
            
        return features
        
    
    def transform(self, waveforms):
        n = self.n_components_by_channel
        all = np.zeros((waveforms.shape[0], waveforms.shape[2]*n), dtype=self.dtype)
        for c, pca in enumerate(self.pcas):
            if pca is None:
                continue
            #~ print(c)
            all[:, c*n:(c+1)*n] = pca.transform(waveforms[:, :, c])
        return all



class GlobalLDA:
    def __init__(self, catalogueconstructor=None, selection=None, **params):

        cc = catalogueconstructor
        
        self.waveforms = cc.get_some_waveforms()
        
        if selection is None:
            #~ waveforms = self.waveforms
            raise NotImplementedError
        else:
            peaks_index, = np.nonzero(selection)
            waveforms = cc.get_some_waveforms(peaks_index=peaks_index)
            labels = cc.all_peaks[peaks_index]['cluster_label']
        
        flatten_waveforms = waveforms.reshape(waveforms.shape[0], -1)
        
        self.lda = sklearn.discriminant_analysis.LinearDiscriminantAnalysis()
        self.lda.fit(flatten_waveforms, labels)
        
        
        #In GlobalPCA all feature represent all channels
        self.channel_to_features = np.ones((cc.nb_channel, self.lda._max_components), dtype='bool')

    def get_features(self, catalogueconstructor):
        features = self.transform(self.waveforms)
        del self.waveforms
        return features


    def transform(self, waveforms):
        flatten_waveforms = waveforms.reshape(waveforms.shape[0], -1)
        return self.lda.transform(flatten_waveforms)



#~ class NeighborhoodPca:
    #~ def __init__(self, waveforms, catalogueconstructor=None, n_components_by_neighborhood=6, radius_um=300., **params):
        #~ cc = catalogueconstructor
        
        #~ self.n_components_by_neighborhood = n_components_by_neighborhood
        #~ self.neighborhood = tools.get_neighborhood(cc.geometry, radius_um)
        
        #~ self.pcas = []
        #~ for c in range(cc.nb_channel):
            #~ neighbors = self.neighborhood[c, :]
            #~ pca = sklearn.decomposition.IncrementalPCA(n_components=n_components_by_neighborhood, **params)
            #~ wfs = waveforms[:,:,neighbors]
            #~ wfs = wfs.reshape(wfs.shape[0], -1)
            #~ pca.fit(wfs)
            #~ self.pcas.append(pca)

        #~ #In full NeighborhoodPca n_components_by_neighborhood feature correspond to one channel
        #~ self.channel_to_features = np.zeros((cc.nb_channel, cc.nb_channel*n_components_by_neighborhood), dtype='bool')
        #~ for c in range(cc.nb_channel):
            #~ self.channel_to_features[c, c*n_components_by_neighborhood:(c+1)*n_components_by_neighborhood] = True

    #~ def transform(self, waveforms):
        #~ n = self.n_components_by_neighborhood
        #~ all = np.zeros((waveforms.shape[0], waveforms.shape[2]*n), dtype=waveforms.dtype)
        #~ for c, pca in enumerate(self.pcas):
            #~ neighbors = self.neighborhood[c, :]
            #~ wfs = waveforms[:,:,neighbors]
            #~ wfs = wfs.reshape(wfs.shape[0], -1)
            #~ all[:, c*n:(c+1)*n] = pca.transform(wfs)
        #~ return all




"""
Some details on parameters for each step.

Preprocessor
---------------------

    * highpass_freq (float): frequency of high pass filter typically 250~500Hz. This remove LFP component in the signal
      Theorically a low value is better (250Hz) but if the signal contain oscillation at high freqyencies 
      (during sleep for insctance) theye must removed so 400Hz should be OK.
    * lowpass_freq (float): low pass frequency (typically 3000~10000Hz) This remove noise in high freuqnecy. This help
       to smooth the spike for peak alignement. This must not exceed niquist frequency (sample_rate/2)
    * smooth_size (int): other possibility to smooth signal. This apply a kernel (more or less triangle) the *smooth_size**
       width in sample. This is like a lowpass filter. If you don't known put 0.
    * common_ref_removal (bool): this substracts sample by sample the median across channels
       When there is a strong noise that appears on all channels (sometimes due to reference) you
       can substract it. This is as if all channels would re referenced numerically to there medians.
    * chunksize (int): the whole processing chain is applied chunk by chunk, this is the chunk size in sample. Typically 1024.
       The smaller size lead to less memory but more CPU comsuption in Peeler. For online, this will be more or less the latency.
    * pad_width (int): size in sample of the margin at the front edge for each chunk to avoid border effect in backward filter.
       In you don't known put None then pad_width will be int(sample_rate/highpass_freq)*3 which is quite robust (<5% error)
       compared to a true offline filtfilt.
    * engine (str): 'numpy' or 'opencl'. There is a double implementation for signal preprocessor : With numpy/scipy
      flavor (and so CPU) or opencl with home made CL kernel (and so use GPU computing). If you have big fat GPU and are able to install
      "opencl driver" (ICD) for your platform the opencl flavor should speedup the peeler because pre processing signal take a quite
      important amoung of time.
    
Peak detector
----------------------

  * peakdetector_engine (str): 'numpy' or 'opencl'.  See signal_preprocessor_engine. Here the speedup is small.
  * peak_sign (str) : sign of the peak ('+' or '-'). The double detection ('+-') is intentionaly NOT implemented is tridesclous
    because it lead to many mistake for users in multi electrode arrays where the same cluster is seen both on negative peak
    and positive rebounce.
  * relative_threshold (str): the threshold without sign with MAD units (robust standard deviation). See :ref:`important_details`.
  * peak_span_ms (float) : this avoid double detection of the same peak in a short span. The units is millisecond.
  
Waveform extraction
--------------------------------

  * wf_left_ms (float); size in ms of the left sweep from the peak index. This number is negative.
  * wf_right_ms (float): size in ms of the right sweep from the peak index. This number is positive.
  * mode (str): 'rand' or 'all' With 'all' all detected peaks are extracted. With 'all' only an randomized subset is taken.
     Note that if you use tridesclous with the script/notebook method you can also choose by yourself which peak are 
     choosen for waveform extraction. This can be usefull to avoid electrical/optical stimlation periods or force peak around
     stimulus periods.


Clean peaks
-------------------

  * alien_value_threshold (float): units=one mad. above this threshold the waveforms is tag as "Alien" and not use for features and clustering
  * mode 'extremum_amplitude' or 'full_waveform': use only the peak value (fast) or the whole waevform (slower)

  * 

    {'name':'extract_waveforms', 'type':'group', 'children' : waveforms_params},
    {'name':'clean_peaks', 'type':'group', 'children' : clean_peaks_params},
    {'name':'peak_sampler', 'type':'group', 'children' : peak_sampler_params},

    {'name': 'alien_value_threshold', 'type': 'float', 'value':100.},
    {'name': 'mode', 'type': 'list', 'values':[, ]}, # full_waveform
]

Peak sampler
--------------------

This step must be carrefully inspected. It select some peak mongs all peaks to make features and clustering.
This highly depend on : the duration on which the catalogue constructor is done + the number of channel (and so the number
of cells) + the density (firing rate) fo each cluster. Since this can't be known in advance, the user must explore cluster and
extract again while changing theses parameters given dense enough clusters.
This have a strong imptact of the CPU and RAM. So do not choose too big numbers.
But you there are too few spikes, cluster could be not detected.
    
  * mode : 'rand', 'rand_by_channel' or 'all' Startegy to select some peak to then make feature and clustering.
         'rand' take a global number of spike indepently of channels detection
         'rand_by_channel' take a random number of spike per channel
         'all' take all spike
         internally with script 'force' allow to select manually which spike we want. For instance to force
         when there is a stimulus.
  * nb_max: when mode='rand'
  * nb_max_by_channel when mode='rand_by_channel'

Noise snippet extraction
--------------------------------------

  * nb_snippet (int):  the number of noise snippet taken in the signal in between peaks.
  

Features extraction
-------------------------------
 
Several methods possible. See :ref:`important_details`.


  * **global_pca**: good option for tetrode.
  
    * n_components (int): number of components of the pca for all the channel.
    

  * **peak_max** : good options when clustering is **sawchaincut**
  
  * **pca_by_channel**: good option for high channel counts

    * n_components_by_channel (int): number of component for each channel.


Cluster
-----------

Several methods possible. See :ref:`important_details`.

  * **kmeans** : `kmeans <http://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html>`_ implemented in sklearn
    
    * n_clusters (int): number of cluster
    
  * **onecluster** no clustering. All label set to 0.
  
  * **gmm** `gaussian mixture model <http://scikit-learn.org/stable/modules/generated/sklearn.mixture.GaussianMixture.html#sklearn.mixture.GaussianMixture>`_ implemented in sklearn 
    
    * n_clusters (int): number of cluster
    * covariance_type (str): 'full', 'tied', 'diag', 'spherical'
    * n_init (int) The number of initializations to perform.
  
  * **agglomerative** `AgglomerativeClustering <http://scikit-learn.org/stable/modules/generated/sklearn.cluster.AgglomerativeClustering.html>`_ implemented in sklearn 
  
    * n_clusters: number of cluster
  
  * **dbscan** `DBSCAN <http://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html>`_ implemented in sklearn
     
    * eps (float): The maximum distance between two samples for them to be considered as in the same neighborhood.

  * **hdbscan** `HDBSCAN <https://hdbscan.readthedocs.io>`_  density base clustering without the problem of the **eps**

  * **isosplit** `ISOSPLIT5 <https://github.com/flatironinstitute/isosplit5>`_ develop for moutainsort (another sorter)

  * **optics** `OPTICS <https://scikit-learn.org/stable/modules/generated/sklearn.cluster.OPTICS.html>`_ implemented in sklearn
     
    * min_samples (int): The number of samples in a neighborhood for a point to be considered as a core point.
  
  * **sawchaincut** Home made automatic clustering, usefull for dense arrays. Autodetect well isolated cluster
    and put to trash ambiguous things.

  * **pruningshears** Another home made automatic clustering. Internaly use hdbscan. Have better performance than **sawcahincut**
    but it is slower.


"""

from collections import OrderedDict
import numpy as np

# TODO copy form .._default_catalogue_params all default value to be consistent

preprocessor_params = [
    {'name': 'highpass_freq', 'type': 'float', 'value':300., 'step': 10., 'suffix': 'Hz', 'siPrefix': True},
    {'name': 'lowpass_freq', 'type': 'float', 'value':5000., 'step': 10., 'suffix': 'Hz', 'siPrefix': True},
    {'name': 'smooth_size', 'type': 'int', 'value':0},
    {'name': 'common_ref_removal', 'type': 'bool', 'value':False},
    #~ {'name': 'chunksize', 'type': 'int', 'value':1024, 'decimals':10},
    {'name': 'pad_width', 'type': 'int', 'value':-1, 'decimals':10, 'limits': (-1, np.inf),},
    {'name': 'engine', 'type': 'list', 'value' : 'numpy', 'values':['numpy', 'opencl']},
]

peak_detector_params = [
    {'name': 'method', 'type': 'list', 'value' : 'global', 'values':['global', 'geometrical']},
    {'name': 'engine', 'type': 'list', 'value' : 'numpy', 'values':['numpy', 'opencl', 'numba']},
    {'name': 'peak_sign', 'type': 'list',  'value':'-', 'values':['-', '+']},
    {'name': 'relative_threshold', 'type': 'float', 'value': 5., 'step': .1,},
    {'name': 'peak_span_ms', 'type': 'float', 'value':0.5, 'step': 0.05, 'suffix': 'ms', 'siPrefix': False},
    {'name':'adjacency_radius_um', 'type': 'float', 'value':200., 'suffix': 'µm', 'siPrefix': False},    
]

waveforms_params = [
    {'name': 'wf_left_ms', 'type': 'float', 'value':-1.0, 'suffix': 'ms', 'step': .1,},
    {'name': 'wf_right_ms', 'type': 'float', 'value': 1.5,  'suffix': 'ms','step': .1,},
    {'name': 'wf_left_long_ms', 'type': 'float', 'value':-2.5, 'suffix': 'ms', 'step': .1,},
    {'name': 'wf_right_long_ms', 'type': 'float', 'value': 3.5,  'suffix': 'ms','step': .1,},
]


clean_peaks_params = [
    {'name': 'alien_value_threshold', 'type': 'float', 'value': np.nan},
    {'name': 'mode', 'type': 'list', 'values':['extremum_amplitude', 'full_waveform']}, # 
]


peak_sampler_params = [
    {'name': 'mode', 'type': 'list', 'values':['rand', 'rand_by_channel', 'all']},
    {'name': 'nb_max', 'type': 'int', 'value':20000},
    {'name': 'nb_max_by_channel', 'type': 'int', 'value':600},
]

#~ clean_waveforms_params =[
    #~ {'name': 'alien_value_threshold', 'type': 'float', 'value':100.},
#~ ]



noise_snippet_params = [
    {'name': 'nb_snippet', 'type': 'int', 'value':300},
]




features_params_by_methods = OrderedDict([
    ('global_pca',  [{'name' : 'n_components', 'type' : 'int', 'value' : 5}]),
    ('peak_max',  []),
    ('pca_by_channel',  [{'name' : 'n_components_by_channel', 'type' : 'int', 'value' : 3},
                                     {'name':'adjacency_radius_um', 'type': 'float', 'value':50., 'suffix': 'µm', 'siPrefix': False},
                                    ]),
    #~ ('neighborhood_pca',  [{'name' : 'n_components_by_neighborhood', 'type' : 'int', 'value' : 3}, 
                                        #~ {'name' : 'radius_um', 'type' : 'float', 'value' : 300., 'step':50.}, 
                                        #~ ]),
])


cluster_params_by_methods = OrderedDict([
    ('pruningshears', [{'name' : 'min_cluster_size', 'type' : 'int', 'value' : 20},
                                {'name':'adjacency_radius_um', 'type': 'float', 'value':50., 'suffix': 'µm', 'siPrefix': False},
                                {'name':'high_adjacency_radius_um', 'type': 'float', 'value':30., 'suffix': 'µm', 'siPrefix': False},
                                ]),
    ('kmeans', [{'name' : 'n_clusters', 'type' : 'int', 'value' : 5}]),
    ('onecluster', []),
    ('gmm', [{'name' : 'n_clusters', 'type' : 'int', 'value' : 5},
                    {'name' : 'covariance_type', 'type' : 'list', 'values' : ['full']},
                    {'name' : 'n_init', 'type' : 'int', 'value' : 10}]),
    ('agglomerative', [{'name' : 'n_clusters', 'type' : 'int', 'value' : 5}]),
    ('dbscan', [{'name' : 'eps', 'type' : 'float', 'value' : 3},
                        {'name' : 'metric', 'type' : 'list', 'values' : ['euclidean', 'l1', 'l2']},
                        {'name' : 'algorithm', 'type' : 'list', 'values' : ['brute', 'auto', 'ball_tree', 'kd_tree', 'brute']},
                    ]),
    ('optics', [{'name' : 'min_samples', 'type' : 'int', 'value' : 5}]),
    ('hdbscan', [{'name' : 'min_cluster_size', 'type' : 'int', 'value' : 20}]),
    ('isosplit5', []),
    ('sawchaincut', [{'name' : 'max_loop', 'type' : 'int', 'value' : 1000},
                                {'name' : 'nb_min', 'type' : 'int', 'value' : 20},
                                {'name' : 'break_nb_remain', 'type' : 'int', 'value' : 30},
                                {'name' : 'kde_bandwith', 'type' : 'float', 'value' : 1., 'step':0.1},
                                {'name' : 'auto_merge_threshold', 'type' : 'float', 'value' : 2., 'step':0.1},
                                {'name':'print_debug', 'type': 'bool', 'value':False},
                            ]),
    
])

clean_cluster_params = [

    {'name':'apply_auto_split', 'type': 'bool', 'value':True},
    
    
    {'name':'apply_trash_not_aligned', 'type': 'bool', 'value':True},
    
    {'name':'apply_auto_merge_cluster', 'type': 'bool', 'value':True},
    
    
    {'name':'apply_trash_low_extremum', 'type': 'bool', 'value':True},
    
    
    {'name':'apply_trash_small_cluster', 'type': 'bool', 'value':True},
    
]


make_catalogue_params = [
    {'name':'inter_sample_oversampling', 'type': 'bool', 'value':False},
    {'name' : 'sparse_thresh_level1', 'type' : 'float', 'value' : 1.5, 'step':0.1},
    {'name' : 'sparse_thresh_level2', 'type' : 'float', 'value' : 3., 'step':0.1},
    
    
]



fullchain_params = [

    # global params
    {'name':'duration', 'type': 'float', 'value':300., 'suffix': 's', 'siPrefix': True},
    {'name': 'chunksize', 'type': 'int', 'value':1024, 'decimals':10},
    {'name' : 'mode', 'type' : 'list', 'values' : ['dense', 'sparse']},
    {'name':'sparse_threshold', 'type': 'float', 'value':1.5},
    {'name' : 'n_spike_for_centroid', 'type' : 'int', 'value' : 350},
    {'name' : 'n_jobs', 'type' : 'int', 'value' : -1},
    
    
    # params for each steps
    {'name':'preprocessor', 'type':'group', 'children': preprocessor_params},
    {'name':'peak_detector', 'type':'group', 'children': peak_detector_params},
    {'name':'noise_snippet', 'type':'group', 'children': noise_snippet_params},
    {'name':'extract_waveforms', 'type':'group', 'children' : waveforms_params},
    {'name':'clean_peaks', 'type':'group', 'children' : clean_peaks_params},
    {'name':'peak_sampler', 'type':'group', 'children' : peak_sampler_params},
    
    #~ {'name':'clean_cluster', 'type': 'bool', 'value':True},
    {'name':'clean_cluster', 'type':'group', 'children' : clean_cluster_params},
    
    {'name':'make_catalogue', 'type':'group', 'children' : make_catalogue_params},
    
]

metrics_params = [
    {'name': 'spike_waveforms_similarity', 'type': 'list', 'values' : [ 'cosine_similarity']},
    {'name': 'cluster_similarity', 'type': 'list', 'values' : [ 'cosine_similarity_with_max']},
    {'name': 'cluster_ratio_similarity', 'type': 'list', 'values' : [ 'cosine_similarity_with_max']},
    {'name': 'size_max', 'type': 'int', 'value':10000000},
]


_common_peeler_params = [
    #~ {'name':'limit_duration', 'type': 'bool', 'value': False},
    {'name': 'chunksize', 'type': 'int', 'value':1024, 'decimals':10},
    {'name':'duration', 'type': 'float', 'value':60., 'suffix': 's', 'siPrefix': True},
    
    
    
    {'name': 'maximum_jitter_shift', 'type': 'int', 'value':4, 'decimals':10},
    {'name':'save_bad_label', 'type': 'bool', 'value':False},
    
]

#~ _classic_peeler_params = _common_peeler_params + [
    #~ {'name': 'argmin_method', 'type': 'list', 'values' : [ 'numpy', 'opencl', 'numba',]},
#~ ]

_geometrical_peeler_params = _common_peeler_params + [
    #~ {'name':'adjacency_radius_um', 'type': 'float', 'value':100., 'suffix': 'µm', 'siPrefix': False},
]

_geometrical_opencl_peeler_params = _common_peeler_params + [
    {'name':'adjacency_radius_um', 'type': 'float', 'value':100., 'suffix': 'µm', 'siPrefix': False},
]


peeler_params_by_methods = OrderedDict([
    #~ ('classic', _classic_peeler_params),
    ('geometrical', _geometrical_peeler_params),
    ('geometrical_opencl', _geometrical_opencl_peeler_params),
])



possible_tags = ['', 'so_bad', 'bad', 'not_so_bad','not_so_good','good', 'so_good', 'better_than_dreams']




